import os
import re
import numpy as np
import healpy as hp
from typing import Any, Optional, Union, Dict
from .routines import _log
from .configurations import Configs
from types import SimpleNamespace
import fnmatch
import sys

def _save_compsep_products(
    config: Configs,
    output_maps: SimpleNamespace,
    compsep_run: Dict[str, Any],
    nsim: Optional[str] = None
) -> None:
    """
    Save component separation products to disk based on the method and simulation parameters.

    Parameters
    ------------
        config: Configs
            Configuration object. It contains paths and parameters for saving outputs.
        output_maps: SimpleNamespace 
            Object containing separated map outputs as attributes.
        compsep_run: Dict 
            Dictionary describing the component separation method and setup.
        nsim: (int, optional)
            Simulation index for saving multiple realizations.

    Returns
    ------------
        None
            It saves the output maps to disk in the specified directory structure 
            based on the component separation method and configuration.
    """

    if 'path_out' not in compsep_run:
        compsep_run["path_out"] = _get_full_path_out(config, compsep_run)

    if compsep_run["method"] in ["cilc", "c_ilc", "mc_cilc","cpilc", "c_pilc"]:
        if 'mixed' in compsep_run["path_out"]:
            with open(os.path.join(compsep_run["path_out"], "constraints_info.txt"), "w") as f: 
                f.write("# Moments and deprojection constraints used in the component separation run\n\n")
                
                moments_header = "# Moments per needlet band (if applicable):\n"
                f.write(moments_header)
                if compsep_run["domain"] == "needlet":
                    for row in compsep_run["constraints"]["moments"]:
                        f.write(" ".join(row) + "\n")
                else:
                    f.write(" ".join(compsep_run["constraints"]["moments"]) + "\n")
                f.write("\n")
                
                depro_header = "# Deprojection coefficients per needlet band (if applicable):\n"
                f.write(depro_header)
                if compsep_run["domain"] == "needlet":
                    for row in compsep_run["constraints"]["deprojection"]:
                        f.write(" ".join(map(str, row)) + "\n")
                else:
                    f.write(" ".join(map(str, compsep_run["constraints"]["deprojection"])) + "\n")
                f.write("\n")
                
                beta_d_header = "# Beta_d values per needlet band (if applicable):\n"
                f.write(beta_d_header)
                if compsep_run["domain"] == "needlet":
                    for row in compsep_run["constraints"]["beta_d"]:
                        f.write(f"{row}\n")
                else:
                    f.write(f'{compsep_run["constraints"]["beta_d"]}\n')
                f.write("\n")
                
                T_d_header = "# T_d values per needlet band (if applicable):\n"
                f.write(T_d_header)
                if compsep_run["domain"] == "needlet":
                    for row in compsep_run["constraints"]["T_d"]:
                        f.write(f"{row}\n")
                else:
                    f.write(f'{compsep_run["constraints"]["T_d"]}\n')
                f.write("\n")
                
                beta_s_header = "# Beta_s values per needlet band (if applicable):\n"
                f.write(beta_s_header)
                if compsep_run["domain"] == "needlet":
                    for row in compsep_run["constraints"]["beta_s"]:
                        f.write(f"{row}\n")
                else:
                    f.write(f'{compsep_run["constraints"]["beta_s"]}\n')
                f.write("\n")

    for attr_name, attr_values in vars(output_maps).items():
        if attr_name == "total":
            label_out = "output_total"
        elif attr_name == "cmb":
            label_out = "output_cmb"
        elif attr_name == "m":
            label_out = "fgd_complexity"
        else:
            label_out = f"{attr_name}_residuals"

        path_c = os.path.join(compsep_run["path_out"], f"{label_out}")
        os.makedirs(path_c, exist_ok=True)

        if compsep_run["method"] in ["gilc","gpilc"]:
            if nsim is not None:
                path_c = os.path.join(path_c, f"{nsim}")
                os.makedirs(path_c, exist_ok=True)

            for f, freq in enumerate(compsep_run["channels_out"]):
                tag = config.instrument.channels_tags[freq]
                filename = (
                    f"{config.field_out}_{label_out}_{tag}_{config.fwhm_out}acm_"
                    f"ns{config.nside}_lmax{config.lmax}"
                )
                if nsim is not None:
                    filename += f"_{nsim}"
                filename += ".fits"

                hp.write_map(os.path.join(path_c, filename), attr_values[f], overwrite=True)

        elif (
            compsep_run["method"] in ["fgd_diagnostic", "fgd_P_diagnostic"]
            and compsep_run["domain"] == "needlet"
        ):
            if nsim is not None:
                path_c = os.path.join(path_c, f"{nsim}")
                os.makedirs(path_c, exist_ok=True)
            
            for j in range(attr_values.shape[-2]):
                filename = (
                    f"{config.field_out}_{label_out}_nl{j}_{config.fwhm_out}acm_"
                    f"ns{config.nside}_lmax{config.lmax}"
                )
                if nsim is not None:
                    filename += f"_{nsim}"
                filename += ".fits"
                if attr_values.ndim == 2:
                    hp.write_map(os.path.join(path_c, filename), attr_values[j], overwrite=True)
                elif attr_values.ndim == 3:
                    hp.write_map(os.path.join(path_c, filename), attr_values[:,j], overwrite=True)
        else:              
            filename = (
                f"{config.field_out}_{label_out}_{config.fwhm_out}acm_"
                f"ns{config.nside}_lmax{config.lmax}"
            )
            if nsim is not None:
                filename += f"_{nsim}"
            filename += ".fits"
            hp.write_map(os.path.join(path_c, filename), attr_values, overwrite=True)
            
def _save_residuals_template(
    config: Configs,
    output_maps: SimpleNamespace,
    compsep_run: Dict[str, Any],
    nsim: Optional[str] = None
) -> None:
    """
    Save residual foreground templates.

    Parameters
    -----------
        config: Configs
            Configuration object. It contains paths and parameters for saving outputs.
        output_maps: SimpleNamespace
            Object containing separated map outputs as attributes.
        compsep_run: Dict
            Dictionary describing the component separation method and setup.
        nsim: str, optional
            Simulation index for saving multiple realizations.

    Returns
    -----------
        None
            It saves the fgd residuals estimate maps to disk in the specified directory structure
            based on the component separation method and configuration.
    
    """
    path_out = os.path.join(config.path_outputs, compsep_run["compsep_path"])

    gnilc_run = (re.search(r'(gilc_[^/]+)', compsep_run["gnilc_path"])).group(1)
    if "needlet" in gnilc_run:
        folder_after = (compsep_run["gnilc_path"]).split(gnilc_run + "/")[1].split("/")[0]
        gnilc_run += f"_{folder_after}"

    for attr_name, attr_values in vars(output_maps).items():
        if attr_name == "total":
            label_out = "fgres_templates"
        elif attr_name == "fgds":
            label_out = "fgres_templates_ideal"
        else:
            label_out = f"fgres_templates_{attr_name}"

        path_c = os.path.join(path_out, f"{label_out}", gnilc_run)
        os.makedirs(path_c, exist_ok=True)

        filename = (
            f"{config.field_out}_{label_out}_{config.fwhm_out}acm_"
            f"ns{config.nside}_lmax{config.lmax}"
        )
        if nsim is not None:
            filename += f"_{nsim}"
        filename += ".fits"

        hp.write_map(os.path.join(path_c, filename), attr_values, overwrite=True)

def _get_full_path_out(config: Configs, compsep_run: Dict[str, Any]) -> str:
    """
    Constructs the full output path for component separation products based on configuration and run options.

    Parameters
    -----------
        config: Configs
            Configuration object.
        compsep_run: dict
            Dictionary containing method and domain setup.

    Returns
    --------
        str
            Full path where outputs should be saved.
    """
    
    if compsep_run["method"] in ["mc_ilc", "mc_cilc", "c_ilc", "c_pilc"]:
        complete_path = f'{compsep_run["method"]}_{compsep_run["domain"]}_bias{compsep_run["ilc_bias"]}_nls{"-".join(map(str, compsep_run["special_nls"]))}' 
    elif compsep_run["method"] == "mcilc":
        complete_path = f'{compsep_run["method"]}_{compsep_run["domain"]}'
    elif compsep_run["method"] in ["gilc", "gpilc"]:
        complete_path = f'{compsep_run["method"]}_{compsep_run["domain"]}_bias{compsep_run["ilc_bias"]}'
        if compsep_run["domain"] == "pixel":
            if compsep_run["m_bias"] != 0:
                complete_path += f"_m{compsep_run['m_bias']:+}"
            if compsep_run["depro_cmb"] is not None:
                complete_path += f"_deproCMB{compsep_run['depro_cmb']}"
        elif compsep_run["domain"] == "needlet":
            m_bias_array = np.array(compsep_run["m_bias"])
            if any(m_bias_array != 0):
                for m_bias in np.unique(m_bias_array[m_bias_array != 0]):
                    nls_bias = np.where(m_bias_array == m_bias)[0]
                    complete_path += f"_m{m_bias:+}_nls{'-'.join(map(str, nls_bias))}"

            depro_array = np.array(compsep_run["depro_cmb"])        
            if any(depro_array != None):
                for depro_val in np.unique(depro_array[depro_array != None]):
                    nls_depro = np.where(depro_array == depro_val)[0]
                    complete_path += f"_deproCMB{depro_val}_nls{'-'.join(map(str, nls_depro))}"
    else:
        complete_path = f'{compsep_run["method"]}_{compsep_run["domain"]}_bias{compsep_run["ilc_bias"]}'

    if compsep_run["method"] in ["gilc", "gpilc", "fgd_diagnostic", "fgd_P_diagnostic"]:
        if not compsep_run["cmb_nuisance"]:
            complete_path += "_nocmbnuisance"

    if compsep_run["method"] != "mcilc":
        if compsep_run["domain"] == "pixel":
            if compsep_run["cov_noise_debias"] != 0.:
                complete_path += f"_noidebias{compsep_run['cov_noise_debias']}"
        elif compsep_run["domain"] == "needlet":
            debias_array = np.array(compsep_run["cov_noise_debias"])        
            if any(debias_array != 0.):
                for debias_val in np.unique(debias_array[debias_array != 0.]):
                    nls_debias = np.where(debias_array == debias_val)[0]
                    complete_path += f"_noidebias{debias_val}_nls{'-'.join(map(str, nls_debias))}"

#    if (config.leakage_correction is not None) and ("QU" in config.field_in) and (config.mask_type == "observed_patch"):
    if (config.leakage_correction is not None) and ("QU" in config.field_in) and (config.mask_observations is not None):
        leak_def = (config.leakage_correction).split("_")[0] + (config.leakage_correction).split("_")[1] 
        if "_recycling" in config.leakage_correction:
            if "_iterations" in config.leakage_correction:
                iterations = int(re.search(r'iterations(\d+)', config.leakage_correction).group(1))
                leak_def += f'_iters{iterations}'
        complete_path += f"_{leak_def}"

    if compsep_run["method"] in ["cilc", "c_ilc", "mc_cilc","cpilc", "c_pilc"]:
        if compsep_run["domain"] == "pixel":
            mom_text = "".join(compsep_run["constraints"]["moments"])
        elif compsep_run["domain"] == "needlet":
            if all(list(set(row)) == list(set(compsep_run["constraints"]["moments"][0])) for row in compsep_run["constraints"]["moments"]):
                mom_text = "".join(compsep_run["constraints"]["moments"][0])
            else:
                mom_text = ""
                for idx, row in enumerate(compsep_run["constraints"]["moments"]):
                    if idx == 0:
                        mom_text += "".join(row)
                    else:
                        if list(set(row)) != list(set(compsep_run["constraints"]["moments"][idx-1])):
                            mom_text += "_" + "".join(row)
        
        if isinstance(compsep_run["constraints"]["deprojection"], float):  # Handle case where deprojection is a single float
            all_depros = [compsep_run["constraints"]["deprojection"]]
        elif isinstance(compsep_run["constraints"]["deprojection"], list):  # Handle case where deprojection is a list
            if all(isinstance(sublist, list) for sublist in compsep_run["constraints"]["deprojection"]):  # Check if it's a list of lists
                all_depros = list(set(element for sublist in compsep_run["constraints"]["deprojection"] for element in sublist))
            else:  # Handle case where deprojection is a flat list
                all_depros = list(set(compsep_run["constraints"]["deprojection"]))
        if len(all_depros)==1:
            if all_depros[0] != 0.:
                mom_text += f"_depro{all_depros[0]}"
        else:
            mom_text += f"_mixeddepro"
            
        if isinstance(compsep_run["constraints"]["beta_d"], float):  # Handle case where beta_d is a single float
            all_betad = [compsep_run["constraints"]["beta_d"]]
        elif isinstance(compsep_run["constraints"]["beta_d"], list):  # Handle case where beta_d is a list
            if all(isinstance(sublist, list) for sublist in compsep_run["constraints"]["beta_d"]):  # Check if it's a list of lists
                all_betad = list(set(element for sublist in compsep_run["constraints"]["beta_d"] for element in sublist))
            else:  # Handle case where beta_d is a flat list
                all_betad = list(set(compsep_run["constraints"]["beta_d"]))
        if len(all_betad)==1:
            if all_betad[0] != 1.54:
                mom_text += f"_bd{all_betad[0]}"
        else:
            mom_text += f"_mixedbd"

        if isinstance(compsep_run["constraints"]["T_d"], float):  # Handle case where T_d is a single float
            all_Td = [compsep_run["constraints"]["T_d"]]
        elif isinstance(compsep_run["constraints"]["T_d"], list):  # Handle case where T_d is a list
            if all(isinstance(sublist, list) for sublist in compsep_run["constraints"]["T_d"]):  # Check if it's a list of lists
                all_Td = list(set(element for sublist in compsep_run["constraints"]["T_d"] for element in sublist))
            else:  # Handle case where T_d is a flat list
                all_Td = list(set(compsep_run["constraints"]["T_d"]))
        if len(all_Td)==1:
            if all_Td[0] != 20.:
                mom_text += f"_Td{all_Td[0]}"
        else:
            mom_text += f"_mixedTd"

        if isinstance(compsep_run["constraints"]["beta_s"], float):  # Handle case where beta_s is a single float
            all_betas = [compsep_run["constraints"]["beta_s"]]
        elif isinstance(compsep_run["constraints"]["beta_s"], list):  # Handle case where beta_s is a list
            if all(isinstance(sublist, list) for sublist in compsep_run["constraints"]["beta_s"]):  # Check if it's a list of lists
                all_betas = list(set(element for sublist in compsep_run["constraints"]["beta_s"] for element in sublist))
            else:  # Handle case where beta_s is a flat list
                all_betas = list(set(compsep_run["constraints"]["beta_s"]))
        if len(all_betas)==1:
            if all_betas[0] != -3.:
                mom_text += f"_bs{all_betas[0]}"
        else:
            mom_text += f"_mixedbs"

        if len(all_betas) > 1 or len(all_depros) > 1 or len(all_Td) > 1 or len(all_betad) > 1:
            case = 1
            while os.path.exists(os.path.join(config.path_outputs, complete_path, f"{mom_text}_case{case}")):
                case += 1
            mom_text += f"_case{case}"

        complete_path = os.path.join(complete_path, mom_text)

    if compsep_run["domain"] == "needlet":
        text_ = f"{compsep_run['needlet_config']['needlet_windows']}"
        if compsep_run["needlet_config"]["needlet_windows"] != "cosine":
            text_ += f'_B{compsep_run["needlet_config"]["width"]}'
            if compsep_run["needlet_config"]["merging_needlets"]:
                merging_needlets = compsep_run["needlet_config"]["merging_needlets"]
                if merging_needlets[0] != 0:
                    merging_needlets.insert(0,0)
                for j_low, j_high in zip(merging_needlets[:-1], merging_needlets[1:]):
                    text_ += f"_j{j_low}j{j_high-1}"
        else:
            for bandpeak in compsep_run["needlet_config"]["ell_peaks"]:
                text_ += f"_{bandpeak}"
        if compsep_run["b_squared"] or compsep_run["method"] in ["pilc", "cpilc", "c_pilc", "gpilc", "fgd_P_diagnostic"]:
            text_ += "_nlsquared"
        complete_path = os.path.join(complete_path, text_)

    if compsep_run["method"] in ["mcilc","mc_ilc","mc_cilc"]:
        text_ = compsep_run["mc_type"]
        for freq_tracer in compsep_run["channels_tracers"]:
            text_ += f"_{config.instrument.channels_tags[freq_tracer]}"
        text_ += f"_{compsep_run['n_patches']}patches"
        complete_path = os.path.join(complete_path, text_)

    path_out = os.path.join(config.path_outputs, complete_path)

    return path_out

def get_gnilc_maps(
    config: Configs,
    path_gnilc: str,
    field_in: Optional[str] = None,
    nsim: Optional[str] = None
) -> SimpleNamespace:
    """
    Load GNILC component separation results (total signal, noise residuals, and foreground residuals) of all frequency channels
    provided in the instrument object and for a given simulation run and field. 

    Parameters
    -----------
        config : Configs
            Configuration object containing instrument and output specifications.
        path_gnilc : str
            Root path to the GNILC output directory. The full path will be given by '{config.path_outputs}/{path_gnilc}'.
        field_in : Optional[str], default=None
            Type of field to load ("T", "QU", "EB", "TQU", "TEB", etc.). If None, default is config.field_out.
        nsim : Optional[Union[str, int]], default=None
            Simulation identifier, if any (used to select specific simulation output files).

    Returns
    --------
        gnilc_maps : SimpleNamespace
            A container with the following attributes:
            - total: np.ndarray of GNILC total signal maps.
            - noise: np.ndarray of noise residual maps (if available).
            - fgds: np.ndarray of foreground residual maps (if available).
    """
    if not os.path.exists(os.path.join(config.path_outputs, path_gnilc)):
        raise ValueError(f"Path {os.path.join(config.path_outputs, path_gnilc)} does not exist.")
    if field_in is None:
        field_in = config.field_out
    
    gnilc_maps = SimpleNamespace()

    if field_in in ["TQU", "TEB"]:
        if config.field_out == "T":
            gnilc_fields = 0
        elif config.field_out in ["QU", "QU_E", "QU_B", "E", "B"]:
            gnilc_fields = (1,2)
        elif config.field_out in ["TQU","TEB"]:
            gnilc_fields = (0,1,2)
    elif field_in in ["QU","EB"]:
        gnilc_fields = (0,1)
    elif field_in in ["T","E","B"]:
        gnilc_fields = 0

    filepath = os.path.join(config.path_outputs, path_gnilc, "output_total")
    if nsim is not None:
        filepath = os.path.join(filepath, nsim)
    if not os.path.exists(filepath):
        raise ValueError(f"Path {filepath} does not contain the expected multifrequency maps.")

    setattr(gnilc_maps, "total", [])

    for f, freq in enumerate(config.instrument.frequency):
        tag = config.instrument.channels_tags[f]
        filename = f"{field_in}_output_total_{tag}_{config.fwhm_out}acm_ns{config.nside}_lmax{config.lmax}"
        if nsim is not None:
            filename += f"_{nsim}"
        filename += ".fits"
        getattr(gnilc_maps, "total").append(hp.read_map(os.path.join(filepath, filename), field=gnilc_fields))
    setattr(gnilc_maps, "total", np.array(getattr(gnilc_maps, "total")))

    filepath = os.path.join(config.path_outputs, path_gnilc, "noise_residuals")
    if nsim is not None:
        filepath = os.path.join(filepath, nsim)
    if not os.path.exists(filepath):
        print(f"Warning: Path {filepath} does not contain the expected noise residuals. Noise debias will not be possible")
    else:
        setattr(gnilc_maps, "noise", [])
        for f, freq in enumerate(config.instrument.frequency):
            tag = config.instrument.channels_tags[f]
            filename = f"{field_in}_noise_residuals_{tag}_{config.fwhm_out}acm_ns{config.nside}_lmax{config.lmax}"
            if nsim is not None:
                filename += f"_{nsim}"
            filename += ".fits"
            getattr(gnilc_maps, "noise").append(hp.read_map(os.path.join(filepath, filename), field=gnilc_fields))
        setattr(gnilc_maps, "noise", np.array(getattr(gnilc_maps, "noise")))

    filepath = os.path.join(config.path_outputs, path_gnilc, "fgds_residuals")
    if nsim is not None:
        filepath = os.path.join(filepath, nsim)
    if os.path.exists(filepath):
        _log(f"Path {filepath} contains the expected foregrounds residuals. The ideal template of foregrounds residuals with no CMB and noise contamination will be computed", verbose=config.verbose)
        setattr(gnilc_maps, "fgds", [])
        for f, freq in enumerate(config.instrument.frequency):
            tag = config.instrument.channels_tags[f]
            filename = f"{field_in}_fgds_residuals_{tag}_{config.fwhm_out}acm_ns{config.nside}_lmax{config.lmax}"
            if nsim is not None:
                filename += f"_{nsim}"
            filename += ".fits"
            getattr(gnilc_maps, "fgds").append(hp.read_map(os.path.join(filepath, filename), field=gnilc_fields))
        setattr(gnilc_maps, "fgds", np.array(getattr(gnilc_maps, "fgds")))

    return gnilc_maps

def save_ilc_weights(
    config: Configs,
    w: np.ndarray,
    compsep_run: Dict,
    nside_: int,
    nl_scale: Optional[Union[int, None]] = None
) -> None:
    """
    Save ILC component separation weights to disk with appropriate metadata in filename.

    Parameters
    ----------
        config : Configs
            Configuration object.
        w : np.ndarray
            component separation weights to be saved.
        compsep_run : dict
            Dictionary with component separation parameters.
        nside_ : int
            HEALPix NSIDE resolution of the output.
        nl_scale : int, optional
            Needlet scale index for the corresponding ILC run.
    
    Returns
    -------
        None
            It saves the weights to disk in the specified directory structure 
            based on the component separation method and configuration.
    """
    path_w = os.path.join(compsep_run["path_out"], "weights")
    os.makedirs(path_w, exist_ok=True)
    filename = os.path.join(path_w, f"weights_{compsep_run['field']}_{config.fwhm_out}acm_ns{nside_}_lmax{config.lmax}")
    if nl_scale is not None:
        filename += f"_nl{nl_scale}"
    if compsep_run["nsim"] is not None:
        filename += f"_{compsep_run['nsim']}"
    np.save(filename, w)

def save_patches(
    config: Configs,
    patches: np.ndarray,
    compsep_run: Dict,
    nl_scale: Optional[Union[int, None]] = None
) -> None:
    """
    Save MC-ILC patches to disk with appropriate metadata in filename.

    Parameters
    ----------
        config : Configs
            Configuration object.
        patches : np.ndarray
            MC-ILC patches to be saved.
        compsep_run : dict
            Dictionary with component separation parameters.
        nl_scale : int, optional
            Needlet scale index for the corresponding ILC run.

    Returns
    -------
        None
            It saves the MC-ILC patches to disk in the specified directory structure
    """
    path_ = os.path.join(compsep_run["path_out"], "patches")
    os.makedirs(path_, exist_ok=True)
    filename = os.path.join(path_, 
            f"patches_{compsep_run['field']}_{config.fwhm_out}acm_ns{config.nside}_lmax{config.lmax}")
    if nl_scale is not None:
        filename += f"_nl{nl_scale}"
    np.save(filename, patches)

def save_spectra(
    config: Configs,
    cls_out: SimpleNamespace,
    compute_cls: Dict[str, Any],
    nsim: Optional[str] = None
) -> None:
    """
    Save the computed spectra to files based on the configuration and compute_cls dictionary.

    Parameters
    ----------
        config : Configs
            Configuration object containing parameters for spectra computation. See `_compute_spectra` for details.
        cls_out : SimpleNamespace
            Object containing computed spectra with attributes for each component.
        compute_cls : dict
            Dictionary containing parameters for spectra computation. See `_compute_spectra` for details.
        nsim : str, optional
            Simulation number to save spectra.

    Returns
    -------
        None
            It saves the computed spectra to disk in the specified directory structure.
    """

    path_spectra = get_path_spectra(config, compute_cls)

    post_filename = f"_{nsim}" if nsim is not None else ""

    pre_filename = "Dls" if config.return_Dell else "Cls"

    for component in compute_cls["components_for_cls"]:
        component_name = component.split('/')[0] if '/' in component else component
        os.makedirs(os.path.join(path_spectra, component), exist_ok=True)
        filename = os.path.join(
            path_spectra,
            f"{component}/{pre_filename}_{config.field_cls_out}_{component_name}_{config.fwhm_out}acm_ns{config.nside}_lmax{config.lmax}{post_filename}.fits"
        )
        hp.write_cl(filename, getattr(cls_out, component_name), overwrite=True)

def get_path_spectra(config: Configs, compute_cls: Dict[str, Any]) -> str:
    """
    Get the path where spectra will be saved based on the compute_cls dictionary.

    Parameters
    ----------
        config : Configs
            Configuration object containing paths and parameters for spectra computation.
        compute_cls : dict
            Dictionary containing parameters for spectra computation, including mask type and fsky.

    Returns
    -------
        str
            Full path to the directory where spectra will be saved.

    """

    path_spectra = os.path.join(compute_cls["path"], 'spectra')
    mask_patterns = ['GAL*+fgres', 'GAL*+fgtemp', 'GAL*+fgtemp^3','GAL*0', 'GAL97', 'GAL99', 'fgres', 'fgtemp', 
            'fgtemp^3', 'config+fgres', 'config+fgtemp', 'config+fgtemp^3', 'config']

    if compute_cls["mask_type"] is None and config.mask_path is None:
        mask_name = 'fullsky'
    elif compute_cls["mask_type"] is None and config.mask_path is not None:
        mask_name = "fullpatch"
    elif any(fnmatch.fnmatch(compute_cls["mask_type"], pattern) for pattern in mask_patterns):
        if 'fgres' in compute_cls["mask_type"] or 'fgtemp' in compute_cls["mask_type"]:
            mask_name = compute_cls["mask_type"] + f"_fsky{compute_cls['fsky']}"
            if "smooth_tracer" in compute_cls:
                mask_name += f"_{compute_cls['smooth_tracer']}deg"
        elif compute_cls["mask_type"] == "config":
            mask_name = "fullpatch"
        else:
            mask_name = compute_cls["mask_type"]
    elif compute_cls["mask_type"] == "from_fits":
        mask_name = compute_cls.setdefault("mask_definition", "masks_from_fits")

    if compute_cls["apodize_mask"] is not None:
        mask_name += f"_apo{compute_cls['apodize_mask']}_{compute_cls['smooth_mask']}deg"

    return os.path.join(path_spectra, mask_name)
    
def _save_mask(mask: np.ndarray, 
               config: Configs, 
               compute_cls: Dict[str, Any], 
               nsim: Optional[str] = None) -> None:
    """
    Save the mask used for power spectra computation.

    Parameters
    ----------
        mask : np.ndarray
            The mask(s) to be saved.
        config : Configs
            Configuration object containing global parameters.
        compute_cls : dict
            Dictionary containing parameters for spectra computation.
        nsim : str, optional
            Simulation number to save the mask.

    Returns
    -------
        None
            It saves the mask to disk in the specified directory structure based on the configuration and compute_cls parameters.
    """
    
    path_mask = get_path_spectra(config, compute_cls)
    
    post_filename = f"_{nsim}" if nsim is not None else ""

    os.makedirs(path_mask, exist_ok=True)
    
    filename = f"mask_{config.field_cls_out}_{config.fwhm_out}acm_ns{config.nside}_lmax{config.lmax}{post_filename}.fits"
    
    hp.write_map(os.path.join(path_mask, filename), mask, overwrite=True)

__all__ = [
    name
    for name, obj in globals().items()
    if callable(obj) and getattr(obj, "__module__", None) == __name__
]
                    




import numpy as np
import healpy as hp
from .configurations import Configs
from .routines import obj_to_array, array_to_obj
from .saving import _save_residuals_template
from .needlets import _get_nside_lmax_from_b_ell, _get_needlet_windows_, _needlet_filtering
from .ilcs import _ilc_post_processing, _get_good_channels_nl
from types import SimpleNamespace
import os
from typing import Union, Dict, Any, List
import sys

def get_residuals_template(
    config: Configs,
    input_alms: SimpleNamespace,
    compsep_run: dict,
    **kwargs
) -> Union[None, SimpleNamespace]:
    """
    Estimates the foreground residuals template from input foregrounds multifrequency templates using 
    component separation weights as defined in compsep_run dictionary.

    Parameters
    ----------
        config : Configs
            Configuration object containing the global settings. It should include the following attributes:
            - nside: The HEALPix resolution parameter.
            - lmax: The maximum multipole for the analysis.
            - fwhm_out: The full width at half maximum for the foreground templates and at which weights
            were computed.
            - pixel_window_out: Boolean indicating whether to apply pixel windowing to the output maps.
            - save_compsep_products: Boolean indicating whether to save the residuals templates.
            - return_compsep_products: Boolean indicating whether to return the residuals templates.
            - mask_type: Type of mask to apply, e.g., "observed_patch".
            - leakage_correction: Method for leakage correction, e.g., "purify", or "recycling". Used 
            only for PILC residuals estimation.
            - field_out: The field to be outputted, e.g., "E", "B", "EB", "QU", "QU_E", or "QU_B".
        
        input_alms : np.ndarray
            Input foregrounds multifrequency templates in alm format. The shape should be (n_freqs, n_fields, n_alms, n_comps) where:
            - n_freqs: Number of frequency channels.
            - n_fields: Number of fields, not None only if multiple fields are being processed simultaneously.
            - n_alms: Number of spherical harmonic coefficients.
            - n_comps: Number of components referring to foreground templates (e.g. total, noise residuals, etc.).
        
        compsep_run : dict
            Dictionary containing the residuals estimates run parameters. It should include:
            - compsep_path: Path to the component separation weights.
            - field_in_cs: The field associated to input templates and component separation weights, e.g., "T", "E", "B", "QU_E", "QU_B".
            - nsim: Simulation number if applicable, otherwise None.
            - adapt_nside: Boolean indicating whether to adapt nside based on the needlet bands.
            - mask: Optional mask to apply to the output maps.
        
        kwargs : dict
            Additional keywords for healpy map2alm function.

    Returns
    -------
        SimpleNamespace or None
            Estimate of foreground residuals for component separation of scalar fields or for PILC.
            It may have the following attributes:
            - fgres_templates: Total foreground residuals.
            - fgres_templates_noise: Noise contamination in the foreground residuals template.
            - fgres_templates_ideal: Pure foreground component in the template.
            Each attribute has shape (n_fields, n_pixels) with n_fields depending on config.field_out.
    """

    templates = _get_fres(config, obj_to_array(input_alms), compsep_run, **kwargs)
    
    if "pilc" not in compsep_run["compsep_path"]:
        templates = _ilc_post_processing(config, templates, compsep_run, **kwargs)

    templates = array_to_obj(templates, input_alms)

    if config.save_compsep_products:
        _save_residuals_template(config, templates, compsep_run, nsim=compsep_run["nsim"])

    if config.return_compsep_products:
        return templates
    return None

def _get_fres(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: dict,
    **kwargs
) -> np.ndarray:
    """
    Estimates the foreground residuals template from input foregrounds multifrequency templates using
    component separation weights as defined in compsep_run dictionary.

    Parameters
    ----------
        config : Configs
            Configuration object containing the global settings. See 'get_residuals_template' for details.
        input_alms : np.ndarray
            Input foregrounds multifrequency templates in alm format. The shape should be (n_freqs, n_fields, n_alms, n_comps).
            n_fields is not None only if multiple fields are being processed simultaneously.
        compsep_run : dict
            Dictionary containing the residuals estimates run parameters. See 'get_residuals_template' for details.
        kwargs : dict
            Additional keywords for healpy map2alm function.
    
    Returns
    -------
        np.ndarray
            Estimate of foreground residuals for component separation of scalar fields.
            The shape will be (n_fields, n_pixels, n_comps) with n_fields depending on config.field_out.
    """

    if "pilc" in compsep_run["compsep_path"]:
        if input_alms.ndim == 4:
            if input_alms.shape[1] == 3:
                raise ValueError("PILC does not support TEB alms, only EB, E or B alms.")
            elif input_alms.shape[1] == 2:
                compsep_run["field"] = "QU"
        elif input_alms.ndim == 3:
            if compsep_run["field_in_cs"] in ["E", "B"]:
                compsep_run["field"] = f'QU_{compsep_run["field_in_cs"]}'
            elif compsep_run["field_in_cs"] in ["QU_E", "QU_B"]:
                compsep_run["field"] = compsep_run["field_in_cs"]
            else:
                raise ValueError(f"Field {compsep_run['field_in_cs']} not supported for PILC residuals estimate.")
        output_maps = _get_fres_P(config, input_alms, compsep_run, **kwargs)
    else:
        if input_alms.ndim == 4:
            fields_ilc = ["T", "E", "B"] if input_alms.shape[1] == 3 else ["E", "B"]
        elif input_alms.ndim == 3:
            if compsep_run["field_in_cs"] in ["T", "E", "B"]:
                fields_ilc = [compsep_run["field_in_cs"]]
            elif compsep_run["field_in_cs"] in ["QU_E", "QU_B"]:
                fields_ilc = [compsep_run["field_in_cs"][-1]]

        if input_alms.ndim == 4:
            output_maps = np.zeros((input_alms.shape[1], 12 * config.nside**2, input_alms.shape[3]))
            for i in range(input_alms.shape[1]):
                compsep_run["field"] = fields_ilc[i]
                output_maps[i] = _get_fres_scalar(config, input_alms[:, i, :, :], compsep_run, **kwargs)
        elif input_alms.ndim == 3:
            compsep_run["field"] = fields_ilc[0]
            output_maps = _get_fres_scalar(config, input_alms, compsep_run, **kwargs)
    
    del compsep_run["field"]

    return output_maps

def _get_fres_scalar(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: dict,
    **kwargs: Any
) -> np.ndarray:
    """
    Estimates the foreground residuals template from input foregrounds multifrequency templates using
    component separation weights as defined in compsep_run dictionary for a single scalar field.

    Parameters
    ----------
        config : Configs
            Configuration object containing the global settings. See 'get_residuals_template' for details.
        input_alms : np.ndarray
            Input foregrounds multifrequency templates in alm format for a single scalar field.
            The shape should be (n_freqs, n_alms, n_comps).
        compsep_run : dict
            Dictionary containing the residuals estimates run parameters. See 'get_residuals_template' for details.
        kwargs : dict
            Additional keywords for healpy map2alm function.
    
    Returns
    -------
        np.ndarray
        Estimate of foreground residuals for component separation of scalar fields in pixel or needlet domain.
        The shape will be (n_pixels, n_comps).
    """

    domain = compsep_run["compsep_path"].split('ilc_')[1].split('_bias')[0]
    if domain == "pixel":
        return _get_fres_pixel(config, input_alms, compsep_run, **kwargs)
    elif domain == "needlet":
        return _get_fres_needlet(config, input_alms, compsep_run, **kwargs)
    else:
        raise ValueError(f"Domain {domain} not supported for foreground residuals estimation.")

def _get_fres_P(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: dict,
    **kwargs: Any
) -> np.ndarray:
    """
    Estimates the foreground residuals template from input foregrounds multifrequency templates using
    component separation PILC component separation weights.

    Parameters
    ----------
        config : Configs
            Configuration object containing the global settings. See 'get_residuals_template' for details.  
        input_alms : np.ndarray
            Input foregrounds multifrequency templates in alm format for polarization field.
            The shape should be (n_freqs, (n_fields), n_alms, n_comps), with n_fields not None if both E and B fields are present.
        compsep_run : dict
            Dictionary containing the residuals estimates run parameters. See 'get_residuals_template' for details.
        kwargs : dict
            Additional keywords for healpy map2alm function.

    Returns
    -------
        np.ndarray
            Estimate of foreground residuals from PILC. Shape is (n_fields, n_pixels, n_comps) where n_fields depends on config.field_out.
    """

    domain = compsep_run["compsep_path"].split('pilc_')[1].split('_bias')[0]
    if domain == "pixel":
        return _get_fres_P_pixel(config, input_alms, compsep_run, **kwargs)
    elif domain == "needlet":
        return _get_fres_P_needlet(config, input_alms, compsep_run, **kwargs)
    else:
        raise ValueError(f"Domain {domain} not supported for PILC foreground residuals estimation.")

def _get_fres_needlet(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: dict,
    **kwargs: Any
) -> np.ndarray:
    """
    Estimates the foreground residuals template from input foregrounds multifrequency templates using
    component separation weights in needlet domain for a single scalar field.

    Parameters
    ----------
        config : Configs
            Configuration object containing the global settings. See 'get_residuals_template' for details.  
        input_alms : np.ndarray
            Input foregrounds multifrequency templates in alm format.
            The shape should be (n_freqs, n_alms, n_comps).
        compsep_run : dict
            Dictionary containing the residuals estimates run parameters. See 'get_residuals_template' for details.
        kwargs : dict
            Additional keywords for healpy map2alm function.
        
    Returns
    -------
        np.ndarray
        Estimate of foreground residuals for component separation in needlet domain.
        Shape is (n_pixels, n_comps).
    """
    nl_filename = os.path.join(config.path_outputs, compsep_run["compsep_path"], "needlet_bands.npy")
    if os.path.exists(nl_filename):
        b_ell = np.load(nl_filename)
    else:
        raise ValueError(f"Needlet bands need to be saved as a npy file in {os.path.join(config.path_outputs, compsep_run['compsep_path'])}")

    output_alms = np.zeros((input_alms.shape[1], input_alms.shape[-1]), dtype=complex)
    for j in range(b_ell.shape[0]):
        output_alms += _get_fres_needlet_j(config, input_alms, compsep_run, b_ell[j], j, **kwargs)
    
    output_maps = np.array([
        hp.alm2map(
            np.ascontiguousarray(output_alms[:, c]),
            config.nside,
            lmax=config.lmax,
            pol=False,
            pixwin=config.pixel_window_out
        ) for c in range(input_alms.shape[-1])
    ]).T
    
    if "mask" in compsep_run:
        output_maps[compsep_run["mask"] == 0.,:] = 0.

    return output_maps
        
def _get_fres_needlet_j(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: dict,
    b_ell: np.ndarray,
    nl_scale: int,
    **kwargs: Any
) -> np.ndarray:
    """
    Estimates the foreground residuals template from input foregrounds multifrequency templates using
    component separation weights for a single needlet band.

    Parameters
    ----------
        config : Configs
            Configuration object containing the global settings. See 'get_residuals_template' for details.  
        input_alms : np.ndarray
            Input foregrounds multifrequency templates in alm format.
            The shape should be (n_freqs, n_alms, n_comps).
        compsep_run : dict
            Dictionary containing the residuals estimates run parameters. See 'get_residuals_template' for details.
        b_ell : np.ndarray
            Needlet band to be used for the residuals estimation.
        nl_scale : int
            Scale of the needlet band, used to determine the weights filename.
        kwargs : dict
            Additional keywords for healpy map2alm function.
    
    Returns
    -------
        np.ndarray
            Estimate of foreground residuals for component separation in needlet domain for a single band.
            Shape is (n_pixels, n_comps).
    """

    if "mask" in compsep_run or not compsep_run["adapt_nside"]:
        nside_, lmax_ = config.nside, config.lmax
    else:
        nside_, lmax_ = _get_nside_lmax_from_b_ell(b_ell,config.nside,config.lmax)
        
    # Load weights
    weights_filename = os.path.join(config.path_outputs,
        compsep_run["compsep_path"],
        f"weights/weights_{compsep_run['field']}_{config.fwhm_out}acm_ns{nside_}_lmax{config.lmax}_nl{nl_scale}"
    )
    if compsep_run["nsim"] is not None:
        weights_filename += f"_{compsep_run['nsim']}"
    weights_filename += ".npy"

    w_ = np.load(weights_filename)

    # Select good channels and apply needlet filtering
    good_channels_nl = _get_good_channels_nl(config, b_ell)

    input_maps_nl = np.zeros((good_channels_nl.shape[0], 12 * nside_**2, input_alms.shape[-1]))
    for n, channel in enumerate(good_channels_nl):
        input_alms_j = _needlet_filtering(input_alms[channel], b_ell, lmax_)
        input_maps_nl[n] = np.array([
            hp.alm2map(np.ascontiguousarray(input_alms_j[:, c]), nside_, lmax=lmax_, pol=False)
            for c in range(input_alms.shape[-1])
        ]).T

    # Apply ILC weights
    if w_.ndim==1:
        output_maps_nl = np.einsum('i,ijk->jk', w_, input_maps_nl)
    elif w_.ndim==2:
        output_maps_nl = np.einsum('ij,ijk->jk', w_, input_maps_nl)

    del input_maps_nl

    # Convert to alms
    output_alms_nl = np.array([
        hp.map2alm(output_maps_nl[:, c], lmax=lmax_, pol=False, **kwargs)
        for c in range(output_maps_nl.shape[-1])
    ]).T

    nl_bands = compsep_run["compsep_path"].rstrip('/').split('/')[-1]

    if "nlsquared" in nl_bands:
        output_alms_j = _needlet_filtering(output_alms_nl, np.ones(lmax_+1), config.lmax)
    else:
        output_alms_j = _needlet_filtering(output_alms_nl, b_ell[:lmax_+1], config.lmax)

    return output_alms_j

def _get_fres_pixel(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: dict,
    **kwargs: Any
) -> np.ndarray:
    """
    Estimates the foreground residuals template from input foregrounds multifrequency templates using
    component separation weights in pixel domain for a single scalar field.

    Parameters
    ----------
        config : Configs
            Configuration object containing the global settings. See 'get_residuals_template' for details.
        input_alms : np.ndarray
            Input foregrounds multifrequency templates in alm format.
            The shape should be (n_freqs, n_alms, n_comps).
        compsep_run : dict
            Dictionary containing the residuals estimates run parameters. See 'get_residuals_template' for details.
        kwargs : dict
            Additional keywords for healpy map2alm function.
    
    Returns
    -------
        np.ndarray
            Estimate of foreground residuals for component separation of a single scalar field in pixel domain.
            Shape is (n_pixels, n_comps).
    """
    good_channels = _get_good_channels_nl(config, np.ones(config.lmax + 1))

    # Convert alms to maps
    input_maps = np.zeros((good_channels.shape[0], 12 * config.nside**2, input_alms.shape[-1]))
    for n, channel in enumerate(good_channels_nl):
        input_maps[n] = np.array([
            hp.alm2map(np.ascontiguousarray(input_alms[channel, :, c]), config.nside, lmax=config.lmax, pol=False)
            for c in range(input_alms.shape[-1])
        ]).T

    # Load weights
    weights_filename = os.path.join(config.path_outputs,
        compsep_run["compsep_path"],
        f"weights/weights_{compsep_run['field']}_{config.fwhm_out}acm_ns{config.nside}_lmax{config.lmax}"
    )
    if compsep_run["nsim"] is not None:
        weights_filename += f"_{compsep_run['nsim']}"
    weights_filename += ".npy"
    w_ = np.load(weights_filename)
    
    # Apply weights
    if w_.ndim==1:
        output_maps = np.einsum('i,ijk->jk', w_, input_maps)
    elif w_.ndim==2:
        output_maps = np.einsum('ij,ijk->jk', w_, input_maps)

    # Optionally apply pixel window function
    if config.pixel_window_out:
        for c in range(output_maps.shape[1]):
            alm_out = hp.map2alm(output_maps[:,c],lmax=config.lmax, pol=False, **kwargs)
            output_maps[:,c] = hp.alm2map(alm_out, config.nside, lmax=config.lmax, pol=False, pixwin=True)

    # Apply mask if available
    if "mask" in compsep_run:
        output_maps[compsep_run["mask"] == 0.,:] = 0.

    return output_maps

def _get_fres_P_needlet(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: dict,
    **kwargs: Any
) -> np.ndarray:
    """
    Estimates the foreground residuals template from input foregrounds multifrequency templates using
    PILC component separation weights in needlet domain.

    Parameters
    ----------
        config : Configs
            Configuration object containing the global settings. See 'get_residuals_template' for details.
        input_alms : np.ndarray
            Input foregrounds multifrequency templates in alm format for polarization field.
            The shape should be (n_freqs, (n_fields), n_alms, n_comps), with n_fields not None if both E and B fields are present.
        compsep_run : dict
            Dictionary containing the residuals estimates run parameters. See 'get_residuals_template' for details.
        kwargs : dict
            Additional keywords for healpy map2alm function.
    
    Returns
    -------
        np.ndarray
            Estimate of foreground residuals from PILC in needlet domain.
            Shape is (n_fields, n_pixels, n_comps) where n_fields depends on config.field_out.
    """
    # Load needlet bands    
    nl_filename = os.path.join(config.path_outputs, compsep_run["compsep_path"], "needlet_bands.npy")
    if os.path.exists(nl_filename):
        b_ell = np.load(nl_filename)
    else:
        raise ValueError(f"Needlet bands need to be saved as a .npy file in {os.path.join(config.path_outputs, compsep_run['compsep_path'])}")

    # Initialize output map
    output_maps = np.zeros((2, hp.nside2npix(config.nside), input_alms.shape[-1]))

    # Loop over needlet bands and compute residuals
    for j in range(b_ell.shape[0]):
        output_maps += _get_fres_P_needlet_j(config, input_alms, compsep_run, b_ell[j], j, **kwargs)
    
    process_outputs = (
        ((config.field_out in ["QU", "QU_E", "QU_B"]) and config.pixel_window_out) or
        (config.field_out in ["E", "B", "EB"])
    )

    if process_outputs:
        for c in range(output_maps.shape[-1]):
#            if "mask" in compsep_run and config.mask_type == "observed_patch" and config.leakage_correction is not None:
#                if "_purify" in config.leakage_correction:
#                    alm_out = purify_master(
#                        output_maps[..., c], compsep_run["mask"], config.lmax,
#                        purify_E=("E" in config.leakage_correction)
#                    )
#                    alm_out = np.concatenate([(0.*alm_out[0])[np.newaxis],alm_out], axis=0)
#                elif "_recycling" in config.leakage_correction:
#                    iterations = (
#                        int(re.search(r'iterations(\d+)', config.leakage_correction).group(1))
#                        if "_iterations" in config.leakage_correction else 0
#                    )
#                    alm_out = purify_recycling(
#                        output_maps[..., c], output_maps[..., 0],
#                        np.ceil(compsep_run["mask"]), config.lmax,
#                        purify_E=("E" in config.leakage_correction),
#                        iterations=iterations
#                    )
#                    alm_out = np.concatenate([(0.*alm_out[0])[np.newaxis],alm_out], axis=0)
#                elif config.leakage_correction=="mask_only":
#                    alm_out = hp.map2alm(np.array([0.*output_maps[0,:,c],output_maps[0,:,c],output_maps[1,:,c]]) * compsep_run["mask"],lmax=config.lmax, pol=True, **kwargs)
#            else:
            alm_out = hp.map2alm(
                [0.*output_maps[0,:,c], output_maps[0,:,c], output_maps[1,:,c]],
                lmax=config.lmax, pol=True, **kwargs
            )

            if (config.field_out in ["QU", "QU_E", "QU_B"]):
                output_maps[...,c] = hp.alm2map(alm_out, config.nside, lmax=config.lmax, pol=True, pixwin=True)[1:]
            elif config.field_out in ["E","B"]:
                output_maps[0,:,c] = hp.alm2map(alm_out[1] if config.field_out=="E" else alm_out[2], config.nside, lmax=config.lmax, pol=False, pixwin=config.pixel_window_out)
            elif config.field_out=="EB":
                output_maps[0,:,c] = hp.alm2map(alm_out[1], config.nside, lmax=config.lmax, pol=False, pixwin=config.pixel_window_out)
                output_maps[1,:,c] = hp.alm2map(alm_out[2], config.nside, lmax=config.lmax, pol=False, pixwin=config.pixel_window_out)

        if config.field_out in ["E","B"]:
            output_maps = output_maps[0]

    # Apply mask if provided
    if "mask" in compsep_run:
        if output_maps.ndim == 3:
            output_maps[:,compsep_run["mask"] == 0.,:] = 0.
        elif output_maps.ndim == 2:
            output_maps[compsep_run["mask"] == 0.,:] = 0.

    return output_maps
        
def _get_fres_P_needlet_j(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: dict,
    b_ell: np.ndarray,
    nl_scale: int,
    **kwargs: Any
) -> np.ndarray:
    """
    Estimates the foreground residuals template from input foregrounds multifrequency templates using
    PILC component separation weights for a single needlet band.

    Parameters
    ----------
        config : Configs
            Configuration object containing the global settings. See 'get_residuals_template' for details.
        input_alms : np.ndarray
            Input foregrounds multifrequency templates in alm format for polarization field.
            The shape should be (n_freqs, (n_fields), n_alms, n_comps), with n_fields not None if both E and B fields are present.
        compsep_run : dict
            Dictionary containing the residuals estimates run parameters. See 'get_residuals_template' for details.
        b_ell : np.ndarray
            Needlet band to be used for the residuals estimation.
        nl_scale : int
            Scale of the needlet band, used to determine the weights filename.
        kwargs : dict
            Additional keywords for healpy map2alm function.
    
    Returns
    -------
        np.ndarray
            Estimate of foreground residuals for component separation in needlet domain for a single band.
            Shape is (n_pixels, n_comps).
    """

    nside_, lmax_ = config.nside, config.lmax
    good_channels_nl = _get_good_channels_nl(config, b_ell)

    # Load PILC weights
    weights_path = os.path.join(config.path_outputs,
        compsep_run["compsep_path"],
        f"weights/weights_{compsep_run['field']}_{config.fwhm_out}acm_ns{nside_}_lmax{config.lmax}_nl{nl_scale}"
    )
    if compsep_run["nsim"] is not None:
        weights_filename += f"_{compsep_run['nsim']}"
    weights_filename += ".npy"
    w_pilc = np.load(weights_filename)

    input_maps_nl = np.zeros((good_channels_nl.shape[0], 2, 12 * nside_**2, input_alms.shape[-1]))
    # Convert alms to filtered Q/U maps
    for n, channel in enumerate(good_channels_nl):
        input_alms_j = np.zeros((2, hp.Alm.getsize(lmax_), input_alms.shape[-1]), dtype=complex)
        if input_alms.ndim == 4:
            for k in range(2):
                input_alms_j[k] = _needlet_filtering(input_alms[channel,k], b_ell, lmax_)
        elif input_alms.ndim == 3:
            if compsep_run["field_in_cs"] in ["QU_E", "E"]:
                input_alms_j[0] = _needlet_filtering(input_alms[channel], b_ell, lmax_)
            elif compsep_run["field_in_cs"] in ["QU_B", "B"]:
                input_alms_j[1] = _needlet_filtering(input_alms[channel], b_ell, lmax_)

        input_alms_j = np.ascontiguousarray(input_alms_j)

        for c in range(input_alms.shape[-1]):
            input_maps_nl[n,...,c] = hp.alm2map(
                np.array([0.*input_alms_j[0, :, c],input_alms_j[0, :, c],input_alms_j[1, :, c]]),
                nside_, lmax=lmax_, pol=True
            )[1:]

    # Apply PILC weights
    if w_pilc.ndim==1:
        output_maps_nl = np.einsum('i,ifjk->fjk', w_pilc, input_maps_nl)
    elif w_pilc.ndim==2:
        if w_pilc.shape[0] == input_maps_nl.shape[0]:
            output_maps_nl = np.einsum('ij,ifjk->fjk', w_pilc, input_maps_nl)
        elif w_pilc.shape[0] == 2:
            q = np.einsum('i,ijk->jk', w_pilc[0], input_maps_nl[:,0]) - np.einsum('i,ijk->jk', w_pilc[1], input_maps_nl[:,1])
            u = np.einsum('i,ijk->jk', w_pilc[1], input_maps_nl[:,0]) + np.einsum('i,ijk->jk', w_pilc[0], input_maps_nl[:,1])
            output_maps_nl = np.array([q, u])
    elif w_pilc.ndim==3:
        q = np.einsum('ij,ijk->jk', w_pilc[0], input_maps_nl[:,0]) - np.einsum('ij,ijk->jk', w_pilc[1], input_maps_nl[:,1])
        u = np.einsum('ij,ijk->jk', w_pilc[1], input_maps_nl[:,0]) + np.einsum('ij,ijk->jk', w_pilc[0], input_maps_nl[:,1])
        output_maps_nl = np.array([q, u])

    del input_maps_nl

    return output_maps_nl

def _get_fres_P_pixel(config: Configs, input_alms, compsep_run, **kwargs):
    """
    Estimates the foreground residuals template from input foregrounds multifrequency templates using 
    component separation PILC component separation weights in pixel domain.

    Parameters
    ----------
        config : Configs
            Configuration object containing the global settings. See 'get_residuals_template' for details.
        input_alms : np.ndarray
            Input foregrounds multifrequency templates in alm format for polarization field.
            The shape should be (n_freqs, (n_fields), n_alms, n_comps), with n_fields not None if both E and B fields are present.
        compsep_run : dict
            Dictionary containing the residuals estimates run parameters. See 'get_residuals_template' for details.
        kwargs : dict
            Additional keywords for healpy map2alm function.
    Returns
    -------
        np.ndarray
            Estimate of foreground residuals from PILC in pixel domain.
            Shape is (n_fields, n_pixels, n_comps) where n_fields depends on config.field_out.
    """
    good_channels = _get_good_channels_nl(config, np.ones(config.lmax + 1))

    input_maps = np.zeros((good_channels.shape[0], 2, 12 * config.nside**2, input_alms.shape[-1]))
    
    for n, channel in enumerate(good_channels):
        for c in range(input_alms.shape[-1]):
            if input_alms.ndim == 4:
                input_maps[n, ..., c] = hp.alm2map(
                    np.array([0. * input_alms[channel, 0, :, c],
                              input_alms[channel, 0, :, c],
                              input_alms[channel, 1, :, c]]),
                    config.nside, lmax=config.lmax, pol=True)[1:]
            elif input_alms.ndim == 3:
                if compsep_run["field_in_cs"] in ["QU_E", "E"]:
                    input_maps[n, ..., c] = hp.alm2map(
                        np.array([0. * input_alms[channel, :, c],
                                  input_alms[channel, :, c],
                                  0. * input_alms[channel, :, c]]),
                        config.nside, lmax=config.lmax, pol=True)[1:]
                elif compsep_run["field_in_cs"] in ["QU_B", "B"]:
                    input_maps[n, ..., c] = hp.alm2map(
                        np.array([0. * input_alms[channel, :, c],
                                  0. * input_alms[channel, :, c],
                                  input_alms[channel, :, c]]),
                        config.nside, lmax=config.lmax, pol=True)[1:]

    weights_filename = os.path.join(config.path_outputs, compsep_run["compsep_path"], "weights")
    weights_filename += f"/weights_{compsep_run['field']}_{config.fwhm_out}acm_ns{config.nside}_lmax{config.lmax}"
    if compsep_run["nsim"] is not None:
        weights_filename += f"_{compsep_run['nsim']}"
    weights_filename += ".npy"
    w_pilc = np.load(weights_filename)
    
    if w_pilc.ndim==1:
        output_maps = np.einsum('i,ifjk->fjk', w_pilc, input_maps)
    elif w_pilc.ndim==2:
        if w_pilc.shape[0] == input_maps.shape[0]:
            output_maps = np.einsum('ij,ifjk->fjk', w_pilc, input_maps)
        elif w_pilc.shape[0] == 2:
            q = np.einsum('i,ijk->jk', w_pilc[0], input_maps[:,0]) - np.einsum('i,ijk->jk', w_pilc[1], input_maps[:,1])
            u = np.einsum('i,ijk->jk', w_pilc[1], input_maps[:,0]) + np.einsum('i,ijk->jk', w_pilc[0], input_maps[:,1])
            output_maps = np.array([q, u])
    elif w_pilc.ndim==3:
        q = np.einsum('ij,ijk->jk', w_pilc[0], input_maps[:,0]) - np.einsum('ij,ijk->jk', w_pilc[1], input_maps[:,1])
        u = np.einsum('ij,ijk->jk', w_pilc[1], input_maps[:,0]) + np.einsum('ij,ijk->jk', w_pilc[0], input_maps[:,1])
        output_maps = np.array([q, u])

    process_outputs = (
        ((config.field_out in ["QU", "QU_E", "QU_B"]) and config.pixel_window_out) or
        (config.field_out in ["E", "B", "EB"])
    )

    if process_outputs:
        for c in range(output_maps.shape[-1]):
#            if "mask" in compsep_run and config.mask_type == "observed_patch" and config.leakage_correction is not None:
#                if "_purify" in config.leakage_correction:
#                    alm_out = purify_master(
#                        output_maps[..., c], compsep_run["mask"], config.lmax,
#                        purify_E=("E" in config.leakage_correction)
#                    )
#                    alm_out = np.concatenate([(0.*alm_out[0])[np.newaxis],alm_out], axis=0)
#                elif "_recycling" in config.leakage_correction:
#                    iterations = (
#                        int(re.search(r'iterations(\d+)', config.leakage_correction).group(1))
#                        if "_iterations" in config.leakage_correction else 0
#                    )
#                    alm_out = purify_recycling(
#                        output_maps[..., c], output_maps[..., 0],
#                        np.ceil(compsep_run["mask"]), config.lmax,
#                        purify_E=("E" in config.leakage_correction),
#                        iterations=iterations
#                    )
#                    alm_out = np.concatenate([(0.*alm_out[0])[np.newaxis],alm_out], axis=0)
                #elif config.leakage_correction=="mask_only":
                #    alm_out = hp.map2alm(np.array([0.*output_maps[0,:,c],output_maps[0,:,c],output_maps[1,:,c]]) * compsep_run["mask"],lmax=config.lmax, pol=True, **kwargs)
#            else:
            alm_out = hp.map2alm(
                [0.*output_maps[0,:,c], output_maps[0,:,c], output_maps[1,:,c]],
                lmax=config.lmax, pol=True, **kwargs
            )

            if (config.field_out in ["QU", "QU_E", "QU_B"]):
                output_maps[...,c] = hp.alm2map(alm_out, config.nside, lmax=config.lmax, pol=True, pixwin=True)[1:]
            elif config.field_out in ["E","B"]:
                output_maps[0,:,c] = hp.alm2map(alm_out[1] if config.field_out=="E" else alm_out[2], config.nside, lmax=config.lmax, pol=False, pixwin=config.pixel_window_out)
            elif config.field_out=="EB":
                output_maps[0,:,c] = hp.alm2map(alm_out[1], config.nside, lmax=config.lmax, pol=False, pixwin=config.pixel_window_out)
                output_maps[1,:,c] = hp.alm2map(alm_out[2], config.nside, lmax=config.lmax, pol=False, pixwin=config.pixel_window_out)

        if config.field_out in ["E","B"]:
            output_maps = output_maps[0]

    if "mask" in compsep_run:
        if output_maps.ndim == 3:
            output_maps[:,compsep_run["mask"] == 0.,:] = 0.
        elif output_maps.ndim == 2:
            output_maps[compsep_run["mask"] == 0.,:] = 0.

    return output_maps

__all__ = [
    name
    for name, obj in globals().items()
    if callable(obj) and getattr(obj, "__module__", None) == __name__
]




"""

# BROOM: Blind Reconstruction Of signals from Observations in the Microwaves

**BROOM** is a Python package for blind component separation and Cosmic Microwave Background (CMB) data analysis.

---

## 📦 Installation

You can install the base package using:

```
pip install cmbroom
```

This installs the core functionality.  
If you plan to use the few functions that depend on `pymaster`, **you must install it separately** (version `>=2.4`).

---

### 🔧 To include `pymaster` automatically:

You can install `cmbroom` along with its optional `pymaster` dependency by running:

```
pip install cmbroom[pymaster]
```

However, `pymaster` requires some additional system libraries to be installed **before** running the above command.

#### ✅ On Ubuntu/Debian:
```
sudo apt update
sudo apt install build-essential python3-dev libfftw3-dev libcfitsio-dev
```

#### ✅ On macOS (using Homebrew):
```
brew install fftw cfitsio
```

### 📦 Dependencies

This package relies on several scientific Python libraries:

- [astropy>=6.0.1](https://www.astropy.org/)
- [numpy>1.18.5](https://numpy.org/)
- [scipy>=1.8](https://scipy.org/)
- [healpy>=1.15](https://healpy.readthedocs.io/)
- [pysm3>=3.3.2](https://pysm3.readthedocs.io/en/latest/#)
- [mtneedlet>=0.0.5](https://javicarron.github.io/mtneedlet/)



"""


from .compsep import (
    component_separation, estimate_residuals, _combine_products, _load_outputs
)
from .simulations import _get_data_foregrounds_, _get_data_simulations_, _get_noise_simulation, _get_cmb_simulation, _get_full_simulations
from .configurations import Configs, get_params
from .spectra import _compute_spectra, _compute_spectra_, _load_cls
from .clusters import get_and_save_real_tracers_B


__version__ = "0.1.1"



import numpy as np
import healpy as hp
from .configurations import Configs
from .routines import _get_local_cov, _EB_to_QU, _E_to_QU, _B_to_QU,\
                      obj_to_array, array_to_obj, _get_bandwidths
from .saving import _save_compsep_products, _get_full_path_out, save_patches, save_ilc_weights
from .needlets import  _get_nside_lmax_from_b_ell, _get_needlet_windows_, _needlet_filtering, _get_good_channels_nl 
from .seds import _get_CMB_SED, _get_moments_SED, _standardize_cilc
from .clusters import _adapt_tracers_path, _cea_partition, _rp_partition, \
                      get_scalar_tracer, get_scalar_tracer_nl, initialize_scalar_tracers

from types import SimpleNamespace
import os
from typing import Any, Optional, Union, Dict, List, Tuple
import sys

def ilc(config: Configs, input_alms: SimpleNamespace, compsep_run: Dict, **kwargs: Any) -> Optional[SimpleNamespace]:
    """
    Perform Internal Linear Combination (ILC) component separation (all its variants) on CMB scalar fields.

    Parameters
    ----------
        config : Configs
            Configuration object containing global settings. It includes:
            - lmax : int, maximum multipole for the component separation.
            - nside : int, HEALPix resolution parameter for compsep products.
            - fwhm_out : float, full width at half maximum of the output beam in arcminutes.
            - pixel_window_out : bool, whether to apply a pixel window to the output maps.
            - field_out : str, desired output fields (e.g., "T", "E", "B", "QU", "TQU", "QU_E", "QU_B").
            - save_compsep_products : bool, whether to save component separation products.
            - return_compsep_products : bool, whether to return component separation products.
            - path_outputs : str, path to save the output files.
            - units : str, units of the output maps (e.g., "uK_CMB"). Used to compute moments, if needed.
            - bandpass_integrate : bool, whether inputs are bandpass-integrated. Used to compute moments, if needed. 
        input_alms : SimpleNamespace
            SimpleNamespace object containing input spherical harmonic coefficients (alms). 
            Each attribute has shape (n_channels, (n_fields), n_alms), where n_fields depend on the fields requested in config.field_out.
        compsep_run : dict
            Dictionary specifying component separation parameters. It includes:
            - "method": Method to use for component separation (e.g., "ilc", "cilc", "mc_ilc").
            - "domain": Domain of the component separation (e.g., "pixel", "needlet").
            - "needlet_config": Dictionary containing needlet settings:
                - "needlet_windows": Type of needlet windows ('cosine', 'standard', 'mexican').
                - "ell_peaks": List of integers defining multipoles of peaks for needlet bands (required for 'cosine').
                - "width": Width of the needlet windows (required for 'standard' and 'mexican').
                - "merging_needlets": Integer or list of integers defining ranges of needlets to be merged.
            - "ilc_bias": Parameter setting the residual percentage ILC bias (default is 0. and covariance will be computed over the whole sky).
            - "reduce_ilc_bias": Boolean. If True, it implements a procedure to attenuate ILC bias.
            - "b_squared": Boolean indicating if the needlet windows should be squared (if domain is "needlet").
            - "adapt_nside": Adapt HEALPix resolution of needlet maps to sampled multipole range. Deafult: False.
            - "save_needlets": boolean. Whether to save adopted needlet bands in the output directory. 
                If not provided, it is set to the value of config.save_compsep_products. Default: True.
            - "save_weights": boolean. Whether to save compsep weights maps in the output directory.
                If not provided, it is set to the value of config.save_compsep_products. Default: True.
            - "mask": (optional) HEALPix mask which will exclude (unobserved) regions from covariance computation. 
                It must be a 1D array with shape (12 * nside**2,). If non-binary, it will be used to weigh pixel in covariance computation.
            - "n_patches": (int) Number of patches to use in MC-ILC. Default: 50.
            - "mc_type": (str), needed if method is "mcilc" or "mc_ilc".
                Type of MC-ILC to use. Options: "cea_ideal", "rp_ideal", "cea_real", "rp_real". Default: "cea_real".
            - "cov_noise_debias": (float, list) Noise covariance debiasing factor. 
                If different from 0. it will debias the covariance matrix by a factor cov_noise_debias * noise_covariance.
                It must be a list with the same length as the number of needlet scales if domain is "needlet", otherwise a single float.
            - "special_nls": (list) List of needlet scales where moment deprojection is applied in c_ilc and clustering in mc_ilc.
            - "constraints": Dictionary to be used if method is "cilc" or "c_ilc". It must contain:
                - "moments": list of strings with moments to be deprojected in cILC. It can be a list of lists if domain is "needlet".
                    Each list will be associated to a needlet band in order.
            Optionally can include:
                - "beta_d": (float, list): dust spectral index to compute dust moments. If list, a different beta_d is used for each needlet band. Default: 1.54.
                - "T_d": (float, list): dust temperature to compute dust moments. If list, a different T_d is used for each needlet band. Default: 19.6.
                - "beta_s": (float, list): synchrotron spectral index to compute synchrotron moments. If list, a different beta_s is used for each needlet band. Default: -3.
                - "deprojection": (float, list): deprojection factor for each moment. If list, a different deprojection factor is used for each moment.
                    It can be a list of lists if domain is "needlet". Each list will be associated to a needlet band in order. Default: 0. (i.e. full deprojection) for each moment and each needlet band.
        **kwargs : dict
            Additional keyword arguments forwarded to hp.map2alm.

    Returns
    -------
        outputs : SimpleNamespace or None
            Output separated components as a SimpleNamespace object if 
            config.return_compsep_products is True; otherwise None.
    """
    # Standardize cilc method parameters if needed
    if compsep_run["method"] in ["cilc", "c_ilc", "mc_cilc"]:
        compsep_run = _standardize_cilc(compsep_run, config.lmax)

    # Check for MC-ILC ideal tracer requirements
    if compsep_run["method"] in ["mcilc", "mc_ilc", "mc_cilc"]:
        if compsep_run["mc_type"] in ["cea_ideal", "rp_ideal"]:
            if not hasattr(input_alms, "fgds"):
                raise ValueError("The input_alms object must have 'fgds' attribute for ideal tracer in MC-ILC.")
    
    if compsep_run["method"] != "mcilc":
        if np.any(np.array(compsep_run["cov_noise_debias"]) != 0.):
            if not hasattr(input_alms, "noise"):
                raise ValueError("The input_alms object must have 'noise'' attribute for debiasing the covariance.")
            compsep_run["noise_idx"] = 2 if hasattr(input_alms, "fgds") else 1
                    
    # Perform the core ILC component separation
    output_maps = _ilc(config, obj_to_array(input_alms), compsep_run, **kwargs)

    compsep_run.pop("noise_idx", None)
    
    # Post-process outputs (e.g., convert EB to QU if needed)
    output_maps = _ilc_post_processing(config, output_maps, compsep_run, **kwargs)

    # Convert array back to object with attributes
    outputs = array_to_obj(output_maps, input_alms)

    del output_maps

    # Save component separation products if requested
    if config.save_compsep_products:
        _save_compsep_products(config, outputs, compsep_run, nsim=compsep_run["nsim"])

    if config.return_compsep_products:
        return outputs

def _ilc_post_processing(
    config: Configs,
    output_maps: np.ndarray,
    compsep_run: Dict,
    **kwargs: Any
) -> np.ndarray:
    """
    Post-process the ILC output maps, including polarization conversions and masking.

    Parameters
    ----------
        config : Configs
            Configuration object containing output settings. See 'ilc' for details.
        output_maps : np.ndarray
            Output maps from the ILC process. 
            Shape is (n_fields, npix, n_components) where n_fields depends on the requested output field in config.field_out.
        compsep_run : dict
            Dictionary specifying component separation parameters. It requires:
            - "mask": (optional) HEALPix mask which will exclude (unobserved) regions from the output maps.
        **kwargs : dict
            Additional keyword arguments forwarded to hp.map2alm.

    Returns
    -------
        np.ndarray
            Post-processed output maps.
    """
    # Handle EB to QU conversion for 3D arrays with appropriate fields
    if output_maps.ndim == 3 and (
        (output_maps.shape[0] == 2 and config.field_out == "QU") or
        (output_maps.shape[0] == 3 and config.field_out == "TQU")
    ):
        outputs = np.zeros_like(output_maps)
        for c in range(output_maps.shape[-1]):
            outputs[:,:,c] = _EB_to_QU(output_maps[:,:,c],config.lmax, **kwargs)
        if "mask" in compsep_run:
            outputs[:,compsep_run["mask"] == 0.,:] = 0.
        return outputs
    # Handle QU_E and QU_B fields for 2D arrays
    elif (output_maps.ndim==2) and (config.field_out in ["QU_E", "QU_B"]):
        output = np.zeros((2, output_maps.shape[0], output_maps.shape[-1]))
        for c in range(output_maps.shape[-1]):
            if config.field_out == "QU_E":
                output[:,:,c] = _E_to_QU(output_maps[:,c],config.lmax, **kwargs)
            elif config.field_out=="QU_B":
                output[:,:,c] = _B_to_QU(output_maps[:,c],config.lmax, **kwargs)
        if "mask" in compsep_run:
            output[:,compsep_run["mask"] == 0.,:] = 0.
        return output
    # For other cases, apply mask if present
    else:
        if "mask" in compsep_run:
            if output_maps.ndim == 3:
                output_maps[:, compsep_run["mask"] == 0., :] = 0.
            elif output_maps.ndim == 2:
                output_maps[compsep_run["mask"] == 0.,:] = 0.
        return output_maps

def _ilc(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: Dict,
    **kwargs: Any
) -> np.ndarray:
    """
    Perform ILC on input alm arrays, either scalar or multi-field.

    Parameters
    ----------
        config : Configs
            Configuration object containing global settings. See 'ilc' for details.
        input_alms : np.ndarray
            Input spherical harmonic coefficients. 
            It must have shape (n_channels, n_fields, n_alms, n_components) for multifield or 
            (n_channels, n_alms, n_components) for scalar fields.
        compsep_run : dict
            Component separation configuration dictionary. See 'ilc' for details.
        **kwargs : dict
            Additional keyword arguments for hp.map2alm.

    Returns
    -------
        np.ndarray
            Output separated maps.
    """
    # Determine fields based on input shape and desired output
    if input_alms.ndim == 4:
        if input_alms.shape[1] == 3:
            fields_ilc = ["T", "E", "B"]
        elif input_alms.shape[1] == 2:
            fields_ilc = ["E", "B"]
    elif input_alms.ndim == 3:
        if config.field_out in ["T", "E", "B"]:
            fields_ilc = [config.field_out]
        elif config.field_out in ["QU_E", "QU_B"]:
            fields_ilc = [config.field_out[-1]]

    # Allocate output array and run ILC compsep per field if needed
    if input_alms.ndim == 4:
        output_maps = np.zeros((input_alms.shape[1], 12 * config.nside**2, input_alms.shape[3]))
        for i in range(input_alms.shape[1]):
            compsep_run["field"] = fields_ilc[i]
            if compsep_run["method"] in ["mcilc", "mc_ilc", "mc_cilc"]:
                input_fgds_alms = np.zeros_like(input_alms[:, i, :, 0]) if "real" in compsep_run["mc_type"] else input_alms[:, i, :, 1]
                compsep_run["tracers"] = initialize_scalar_tracers(
                    config, input_fgds_alms, compsep_run, field=compsep_run["field"], **kwargs
                )
                del input_fgds_alms
            output_maps[i] = _ilc_scalar(config, input_alms[:, i, :, :], compsep_run, **kwargs)

    elif input_alms.ndim == 3:
        compsep_run["field"] = fields_ilc[0]
        if compsep_run["method"] in ["mcilc", "mc_ilc", "mc_cilc"]:
            input_fgds_alms = np.zeros_like(input_alms[...,0]) if "real" in compsep_run["mc_type"] else input_alms[...,1]
            compsep_run["tracers"] = initialize_scalar_tracers(config, input_fgds_alms, compsep_run, field=compsep_run["field"], **kwargs)
            del input_fgds_alms
        output_maps = _ilc_scalar(config, input_alms, compsep_run, **kwargs)
    
    del compsep_run["field"]

    return output_maps

def _ilc_scalar(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: Dict,
    **kwargs: Any
) -> np.ndarray:
    """
    Execute ILC in the specified domain (pixel, needlet) on scalar field maps.

    Parameters
    ----------
        config : Configs
            Configuration object with global parameters. See 'ilc' for details.
        input_alms : np.ndarray
            Input alm array for a single scalar field. It must have shape (n_channels, n_alms, n_components).
        compsep_run : dict
            Component separation parameters. See 'ilc' for details.
        **kwargs : dict
            Additional keyword arguments passed internally.

    Returns
    -------
        np.ndarray
            Output maps after ILC processing.
    """
    if compsep_run["domain"] == "pixel":
        return _ilc_pixel(config, input_alms, compsep_run, **kwargs)
    elif compsep_run["domain"] == "needlet":
        return _ilc_needlet(config, input_alms, compsep_run, **kwargs)
#    elif compsep_run["domain"] == "harmonic":
#        output_maps = _hilc(config, input_alms, compsep_run)
    else:
        raise ValueError(f"Unsupported domain {compsep_run['domain']} in compsep.")

def _ilc_needlet(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: Dict,
    **kwargs: Any
) -> np.ndarray:
    """
    Perform ILC in the needlet domain.

    Parameters
    ----------
        config : Configs
            Configuration object with global settings. See 'ilc' for details.
        input_alms : np.ndarray
            Input alm array. It must have shape (n_channels, n_alms, n_components).
        compsep_run : dict
            Component separation parameters including needlet configuration. See 'ilc' for details.
        **kwargs : dict
            Additional keyword arguments for hp.map2alm.

    Returns
    -------
        np.ndarray
            Output maps after needlet domain ILC.
    """
    b_ell = _get_needlet_windows_(compsep_run["needlet_config"], config.lmax)
    if compsep_run["b_squared"]:
        b_ell = b_ell**2

    if compsep_run['save_needlets']:
        compsep_run["path_out"] = _get_full_path_out(config, compsep_run)
        os.makedirs(compsep_run["path_out"], exist_ok=True)
        np.save(os.path.join(compsep_run["path_out"], "needlet_bands"), b_ell)

    output_alms = np.zeros((input_alms.shape[1], input_alms.shape[-1]), dtype=complex)
    for j in range(b_ell.shape[0]):
        output_alms += _ilc_needlet_j(config, input_alms, compsep_run, b_ell[j], j, **kwargs)
    
    output_maps = np.array(
        [hp.alm2map(np.ascontiguousarray(output_alms[:, c]), config.nside, lmax=config.lmax, pol=False, pixwin=config.pixel_window_out)
            for c in range(input_alms.shape[-1])]).T
    
    if "mask" in compsep_run:
        output_maps[compsep_run["mask"] == 0.,:] = 0.

    return output_maps
        
def _ilc_needlet_j(
    config: Configs,
    input_alms: np.ndarray,
    compsep_run: Dict,
    b_ell: np.ndarray,
    nl_scale: int,
    **kwargs: Any
) -> np.ndarray:
    """
    Compute the needlet-scale ILC component separated maps.

    Parameters
    ----------
        config : Configs
            Configuration object. See 'ilc' for details.
        input_alms : np.ndarray
            Input alm array. Shape is (n_channels, n_alms, n_components).
        compsep_run : dict
            Component separation configuration. See 'ilc' for details.
        b_ell : np.ndarray
            Needlet window function. Shape is (lmax+1,).
        nl_scale : int
            Current needlet scale index.
        **kwargs : dict
            Additional keyword arguments.

    Returns
    -------
        np.ndarray
            Output alm array at the requested needlet scale.
    """
    # Determine nside and lmax for this scale
    if "mask" in compsep_run:
        nside_, lmax_ = config.nside, config.lmax
    else:
        if compsep_run["adapt_nside"]:
            nside_, lmax_ = _get_nside_lmax_from_b_ell(b_ell,config.nside,config.lmax)
        else:
            nside_, lmax_ = config.nside, config.lmax
    
    # Get frequency channels to be adopted in component separation at this scale
    good_channels_nl = _get_good_channels_nl(config, b_ell)

    input_maps_nl = np.zeros((good_channels_nl.shape[0], 12 * nside_**2, input_alms.shape[-1]))
    for n, channel in enumerate(good_channels_nl):
        input_alms_j = _needlet_filtering(input_alms[channel], b_ell, lmax_)
        input_maps_nl[n] = np.array(
            [hp.alm2map(np.ascontiguousarray(input_alms_j[:, c]), nside_, lmax=lmax_, pol=False)
                for c in range(input_alms.shape[-1])]).T

    # Run either MC-ILC or ILC
    if (compsep_run["method"]=="mcilc") or ((compsep_run["method"]=="mc_ilc" or compsep_run["method"]=="mc_cilc") and nl_scale in compsep_run["special_nls"]):
        tracer_nl = get_scalar_tracer_nl(compsep_run["tracers"], nside_, lmax_, b_ell)
        output_maps_nl = _mcilc_maps(config, input_maps_nl, tracer_nl, compsep_run, b_ell, nl_scale=nl_scale)
    else:
        output_maps_nl = _ilc_maps(config, input_maps_nl, compsep_run, b_ell, nl_scale=nl_scale)

    del input_maps_nl

    output_alms_nl = np.array(
        [hp.map2alm(output_maps_nl[:, c], lmax=lmax_, pol=False, **kwargs) for c in range(output_maps_nl.shape[-1])]
    ).T

    if compsep_run["b_squared"]:
        output_alms_j = _needlet_filtering(output_alms_nl, np.ones(lmax_+1), config.lmax)
    else:
        output_alms_j = _needlet_filtering(output_alms_nl, b_ell[:lmax_+1], config.lmax)
        
    return output_alms_j

def _ilc_pixel(config: Configs, input_alms: np.ndarray, compsep_run: Dict, **kwargs) -> np.ndarray:
    """
    Computes ILC maps from input alms in pixel space using the specified component separation method.

    Parameters
    ----------
        config : Configs 
            Configuration parameters. See 'ilc' for details.
        input_alms : np.ndarray
            Input spherical harmonic coefficients of shape (n_channels, n_alm, n_components).
        compsep_run : dict
            Dictionary specifying the component separation configuration. See 'ilc' for details.
        **kwargs : dict
            Additional keyword arguments for hp.map2alm.

    Returns
    -------
        np.ndarray
            Reconstructed sky maps in pixel space.
    """
    npix = 12 * config.nside ** 2
    _, _, n_comps = input_alms.shape
    
    good_channels = _get_good_channels_nl(config, np.ones(config.lmax + 1))
    input_maps = np.zeros((good_channels.shape[0], npix, n_comps))
    
    for n, channel in enumerate(good_channels):
        input_maps[n] = np.array([
            hp.alm2map(np.ascontiguousarray(input_alms[channel, :, c]), config.nside,
                       lmax=config.lmax, pol=False) for c in range(n_comps)
        ]).T


    if compsep_run["method"]=="mcilc":
        tracer = get_scalar_tracer(compsep_run["tracers"])
        output_maps = _mcilc_maps(config, input_maps, tracer, compsep_run, np.ones(config.lmax+1))    
    else:
        output_maps = _ilc_maps(config, input_maps, compsep_run, np.ones(config.lmax+1))
    
    if config.pixel_window_out:
        for c in range(output_maps.shape[1]):
            alm_out = hp.map2alm(output_maps[:,c],lmax=config.lmax, pol=False, **kwargs)
            output_maps[:, c] = hp.alm2map(alm_out, config.nside, lmax=config.lmax,
                                           pol=False, pixwin=True)

    if "mask" in compsep_run:
        output_maps[compsep_run["mask"] == 0.,:] = 0.

    return output_maps

def _ilc_maps(config: Configs, input_maps: np.ndarray, compsep_run: Dict,
              b_ell: np.ndarray, nl_scale: Optional[Union[int, None]] = None) -> np.ndarray:
    """
    Computes ILC weights and outputs the cleaned map using ILC methods with inclusion of moments deprojection (if needed).

    Parameters
    ----------
        config : Configs
            Configuration object containing global settings. See 'ilc' for details.
        input_maps : np.ndarray
            Input maps of shape (n_channels, n_pixels, n_components).
        compsep_run : dict
            Dictionary with ILC run configuration. See 'ilc' for details.
        b_ell : np.ndarray
            Needlet harmonic window function. Used if `compsep_run["domain"]` is "needlet".
            If `compsep_run["domain"]` is "pixel", it should be an array of ones with shape (lmax+1,).
        nl_scale : int, optional
            Needlet scale index corresponding to the current ILC run. Used for:
            - Deprojection of moments in cILC.
            - Save weights with proper label.

    Returns
    -------
        np.ndarray
            Cleaned output sky maps of shape (n_pixels, n_components).
    """

    good_channels = _get_good_channels_nl(config, b_ell)
    freqs = np.array(config.instrument.frequency)[good_channels]

    bandwidths = _get_bandwidths(config, good_channels)

    A_cmb = _get_CMB_SED(freqs, units=config.units, bandwidths=bandwidths)
        
    if compsep_run["method"] == "cilc":
        if nl_scale is None:
            compsep_run["A"] = _get_moments_SED(freqs, compsep_run["constraints"]["moments"], beta_d=compsep_run["constraints"]["beta_d"], T_d=compsep_run["constraints"]["T_d"], beta_s=compsep_run["constraints"]["beta_s"], units=config.units, bandwidths=bandwidths)
            compsep_run["e"] = np.array(compsep_run["constraints"]["deprojection"])
        else:
            compsep_run["A"] = _get_moments_SED(freqs, compsep_run["constraints"]["moments"][nl_scale], beta_d=compsep_run["constraints"]["beta_d"][nl_scale], T_d=compsep_run["constraints"]["T_d"][nl_scale], beta_s=compsep_run["constraints"]["beta_s"][nl_scale], units=config.units, bandwidths=bandwidths)
            compsep_run["e"] = np.array(compsep_run["constraints"]["deprojection"][nl_scale])

    elif (compsep_run["method"] == "c_ilc") and (nl_scale is not None) and (nl_scale in compsep_run["special_nls"]):
        compsep_run["A"] = _get_moments_SED(freqs, compsep_run["constraints"]["moments"][compsep_run["special_nls"] == nl_scale], beta_d=compsep_run["constraints"]["beta_d"][compsep_run["special_nls"] == nl_scale], T_d=compsep_run["constraints"]["T_d"][compsep_run["special_nls"] == nl_scale], beta_s=compsep_run["constraints"]["beta_s"][compsep_run["special_nls"] == nl_scale], units=config.units, bandwidths=bandwidths)
        compsep_run["e"] = np.array(compsep_run["constraints"]["deprojection"][compsep_run["special_nls"] == nl_scale])

    elif (compsep_run["method"] == "mc_cilc") and (nl_scale is not None) and (nl_scale not in compsep_run["special_nls"]):
        nl_scale_cilc = nl_scale-len(np.array(compsep_run["special_nls"])[np.array(compsep_run["special_nls"]) < 2])
        compsep_run["A"] = _get_moments_SED(freqs, compsep_run["constraints"]["moments"][nl_scale_cilc], beta_d=compsep_run["constraints"]["beta_d"][nl_scale_cilc], T_d=compsep_run["constraints"]["T_d"][nl_scale_cilc], beta_s=compsep_run["constraints"]["beta_s"][nl_scale_cilc], units=config.units, bandwidths=bandwidths)
        compsep_run["e"] = np.array(compsep_run["constraints"]["deprojection"][nl_scale_cilc])

    cov = get_ilc_cov(input_maps[...,0], config.lmax, compsep_run, b_ell)
    noise_debias = compsep_run["cov_noise_debias"] if compsep_run["domain"] == "pixel" else compsep_run["cov_noise_debias"][nl_scale]
    if noise_debias != 0.:
        cov_n = get_ilc_cov(input_maps[...,compsep_run["noise_idx"]], config.lmax, compsep_run, b_ell)
        cov = cov - noise_debias * cov_n
        del cov_n

    inv_cov = get_inv_cov(cov)
    del cov

    w_ilc = get_ilc_weights(A_cmb, inv_cov, input_maps.shape, compsep_run)
    del inv_cov

    if compsep_run["save_weights"]:
        if 'path_out' not in compsep_run:
            compsep_run["path_out"] = _get_full_path_out(config, compsep_run)
        save_ilc_weights(config, w_ilc, compsep_run,
                         hp.npix2nside(input_maps.shape[-2]), nl_scale=nl_scale)
    
    compsep_run.pop("A", None)
    compsep_run.pop("e", None)

    if compsep_run["ilc_bias"] == 0.:
        output_maps = np.einsum('i,ijk->jk', w_ilc, input_maps)
    else:
        output_maps = np.einsum('ij,ijk->jk', w_ilc, input_maps)

    return output_maps

def _mcilc_maps(config: Configs, input_maps: np.ndarray, tracer: np.ndarray,
                compsep_run: Dict, b_ell: np.ndarray,
                nl_scale: Optional[Union[int, None]] = None) -> np.ndarray:
    """
    Computes foreground-cleaned sky maps using the MC-ILC method.

    Depending on the specified `mc_type`, it delegates to either the CEA-based or 
    random partitioning (RP) implementation.

    Parameters
    ----------
        config : Configs
            Configuration object containing instrument, resolution, and output settings. See 'ilc' for details.
        input_maps : np.ndarray
            Input sky maps of shape (n_channels, n_pixels, n_components).
        tracer : np.ndarray
            Tracer map used to define spatial partitions for independent component separation. 
            It should have shape (n_pixels,).
        compsep_run : dict
            Dictionary specifying the MCILC method and associated parameters:
            - "mc_type": Type of MCILC ('cea' or 'rp').
            - "n_patches": Number of spatial patches.
            - Optional flags such as "save_weights".
            See 'ilc' for details.
        b_ell : np.ndarray
            Needlet harmonic window function. Shape is (lmax+1,).
            Used if `compsep_run["domain"]` is "needlet".
            If `compsep_run["domain"]` is "pixel", it should be an array of ones with shape (lmax+1,).
        nl_scale : int, optional
            Needlet scale index for the current ILC run. Used for:
            - saving weights with proper label.

    Returns
    -------
        np.ndarray
            Cleaned output sky maps of shape (n_pixels, n_components).
    """
    good_channels = _get_good_channels_nl(config, b_ell)
    bandwidths = _get_bandwidths(config, good_channels)

    # Get CMB SED for the good channels
    A_cmb = _get_CMB_SED(np.array(config.instrument.frequency)[good_channels], units=config.units, bandwidths=bandwidths)

    if "cea" in compsep_run["mc_type"]:
        return _mcilc_cea_(config, input_maps, tracer, compsep_run, A_cmb, nl_scale=nl_scale)
    elif "rp" in compsep_run["mc_type"]:
        return _mcilc_rp_(config, input_maps, tracer, compsep_run, A_cmb, nl_scale=nl_scale)

def _mcilc_cea_(config: Configs, input_maps: np.ndarray, tracer: np.ndarray,
                compsep_run: Dict, A_cmb: np.ndarray,
                nl_scale: Optional[Union[int, None]] = None) -> np.ndarray:
    """
    Performs MC-ILC CMB reconstruction using CEA (Cluster of Equal Area) partitioning.

    The sky is divided into non-overlapping spatial patches based on a tracer map and
    the Healpix grid, allowing region-specific ILC weight estimation.

    Parameters
    ----------
        config : Configs
            Configuration object with all global settings. See 'ilc' for details.
        input_maps : np.ndarray
            Input sky maps with shape (n_channels, n_pixels, n_components).
        tracer : np.ndarray
            Tracer map that defines sky partition.
        compsep_run : dict
            Component separation configuration, including:
            - "n_patches": Number of CEA patches.
            - "save_weights": Flag to save weights.
            See 'ilc' for details.
        A_cmb : np.ndarray
            Spectral energy distribution (SED) vector for the CMB.
        nl_scale : int, optional
            Needlet scale index for the current ILC run. Used for:
            - saving weights with proper label.

    Returns
    -------
        np.ndarray
            Output sky map cleaned via region-specific ILC weights, shape (n_pixels, n_components).
    """

    mask_mcilc = compsep_run.get("mask", np.ones(input_maps.shape[-2]))

    patches = _cea_partition(tracer, compsep_run["n_patches"], mask=mask_mcilc)
    if compsep_run["save_patches"] and (compsep_run['nsim'] is None or int(compsep_run['nsim']) == config.nsim_start):
        if 'path_out' not in compsep_run:
            compsep_run["path_out"] = _get_full_path_out(config, compsep_run)
        save_patches(config, patches, compsep_run, nl_scale=nl_scale)

    w_mcilc = get_mcilc_weights(input_maps[...,0], patches, A_cmb, compsep_run)

    if compsep_run["save_weights"]:
        if 'path_out' not in compsep_run:
            compsep_run["path_out"] = _get_full_path_out(config, compsep_run)
        save_ilc_weights(config, w_mcilc, compsep_run, hp.npix2nside(input_maps.shape[-2]), nl_scale=nl_scale)
    
    compsep_run.pop("A", None)
    compsep_run.pop("e", None)

    return np.einsum('ij,ijk->jk', w_mcilc, input_maps)
    
def _mcilc_rp_(config: Configs, input_maps: np.ndarray, tracer: np.ndarray,
               compsep_run: Dict, A_cmb: np.ndarray,
               iterations: int = 30, nl_scale: Optional[Union[int, None]] = None) -> np.ndarray:
    """
    Performs MCILC reconstruction using Random Partitioning (RP) of the sky.

    The tracer map is used to generate multiple random spatial partitions. ILC weights
    are estimated independently for each realization and averaged across iterations
    to reduce bias and improve robustness.

    Parameters
    ----------
        config : Configs
            Configuration object including instrument and output settings. See 'ilc' for details.
        input_maps : np.ndarray
            Input sky maps with shape (n_channels, n_pixels, n_components).
        tracer : np.ndarray
            Tracer map used to guide random partitions.
        compsep_run : dict
            Dictionary with MCILC-specific parameters:
            - "n_patches": Number of partitions per iteration.
            - "save_weights": Whether to save averaged ILC weights.
            See 'ilc' for details.
        A_cmb : np.ndarray
            Spectral energy distribution (SED) of the CMB.
        iterations : int, default=30
            Number of random partition iterations for averaging.
        nl_scale : int or str, optional
            Needlet scale index for the current ILC run. Used for:
            - saving weights with proper label.

    Returns
    -------
        np.ndarray
            Cleaned sky map averaged over random partition realizations,
            shape (n_pixels, n_components).
    """
    output_maps = np.zeros((input_maps.shape[1], input_maps.shape[-1]))
    mask_mcilc = compsep_run.get("mask", np.ones(input_maps.shape[-2]))

    do_save_patches = compsep_run["save_patches"] and (compsep_run['nsim'] is None or int(compsep_run['nsim']) == config.nsim_start)

    if do_save_patches:
        patches_set = []

    for it in range(iterations):  
        patches = _rp_partition(tracer, compsep_run["n_patches"], mask=mask_mcilc)
        if do_save_patches:
            patches_set.append(patches)

        w_mcilc = get_mcilc_weights(input_maps[...,0], patches, A_cmb, compsep_run)
        if compsep_run["save_weights"]:
            if it == 0:
                w_mcilc_save = np.copy(w_mcilc) / iterations
            else:
                w_mcilc_save += w_mcilc / iterations
        output_maps += (np.einsum('ij,ijk->jk', w_mcilc, input_maps) / iterations)
        
    if do_save_patches:
        if 'path_out' not in compsep_run:
            compsep_run["path_out"] = _get_full_path_out(config, compsep_run)
        save_patches(config, np.array(patches_set), compsep_run, nl_scale=nl_scale)

    if compsep_run["save_weights"]:
        if 'path_out' not in compsep_run:
            compsep_run["path_out"] = _get_full_path_out(config, compsep_run)
        save_ilc_weights(config, w_mcilc_save, compsep_run,
                         hp.npix2nside(input_maps.shape[-2]), nl_scale=nl_scale)

    compsep_run.pop("A", None)
    compsep_run.pop("e", None)
    
    return output_maps

def get_ilc_weights(
    A_cmb: np.ndarray,
    inv_cov: np.ndarray,
    input_shapes: Tuple[int, ...],
    compsep_run: Dict
) -> np.ndarray:
    """
    Compute Internal Linear Combination (ILC) weights for a given spectral component (typically CMB).

    Parameters
    ----------
        A_cmb : np.ndarray
            Spectral response vector of the target component (e.g., CMB), shape (n_channels,).
        inv_cov : np.ndarray
            Inverse of the covariance matrix. Can be 2D or 3D depending on bias settings.
        input_shapes : tuple
            Shape of the input map, used to reshape or broadcast weights.
        compsep_run : dict
            Dictionary specifying component separation settings, including:
            - 'A': Optional constraint matrix for deprojection of additional components. Shape (n_components, n_channels).
            - 'e': Deprojection vector. Shape (n_components,).
            - 'ilc_bias': Flag indicating if covariance is pixel-independent (0.0) or not.
            - 'mask': Optional binary mask used for excluding unobserved regions.

    Returns
    -------
        np.ndarray
            ILC weights with shape (n_channels, n_pixels) if ilc_bias is not 0.0,
            otherwise (n_channels,).
    """
    if "A" in compsep_run:
        compsep_run["A"] = np.vstack((A_cmb, compsep_run["A"]))
        compsep_run["e"] = np.insert(compsep_run["e"], 0, 1.)
        if compsep_run["ilc_bias"] == 0.:
            inv_ACA = np.linalg.inv(
                np.einsum("zi,il->zl", compsep_run["A"], np.einsum("ij,lj->il", inv_cov, compsep_run["A"]))
            )
            w_ilc = np.einsum(
                "l,lj->j", compsep_run["e"],
                np.einsum("lz,zj->lj", inv_ACA, np.einsum("zi,ij->zj", compsep_run["A"], inv_cov))
            )
            del inv_ACA
        else:
            inv_ACA = np.linalg.inv(
                np.einsum("zi,ilk->zlk", compsep_run["A"], np.einsum("ijk,lj->ilk", inv_cov, compsep_run["A"])).T
            ).T
            w_ilc=np.zeros((input_shapes[0],input_shapes[-2]))
            w_ = np.einsum(
                "l,ljk->jk", compsep_run["e"],
                np.einsum("lzk,zjk->ljk", inv_ACA, np.einsum("zi,ijk->zjk", compsep_run["A"], inv_cov))
            )
            for i in range(input_shapes[0]):
                if "mask" in compsep_run:
                    w_ilc[i, compsep_run["mask"] > 0.] = w_[i]
                else:
                    w_ilc[i]=hp.ud_grade(w_[i],hp.npix2nside(input_shapes[-2]))
            del w_, inv_ACA
    else:
        if compsep_run["ilc_bias"] == 0.:
            w_ilc = (A_cmb.T @ inv_cov) / (A_cmb.T @ inv_cov @ A_cmb) 
        else:
            w_ilc=np.zeros((input_shapes[0],input_shapes[-2]))
            AT_invC = np.einsum('j,ijk->ik', A_cmb, inv_cov) # np.sum(inv_cov,axis=1)
            AT_invC_A = np.einsum('j,ijk, i->k', A_cmb, inv_cov, A_cmb) #np.sum(inv_cov,axis=(0,1))
            for i in range(input_shapes[0]):
                if "mask" in compsep_run:
                    w_ilc[i, compsep_run["mask"] > 0.] = AT_invC[i]/AT_invC_A
                else:
                    w_ilc[i]=hp.ud_grade(AT_invC[i]/AT_invC_A,hp.npix2nside(input_shapes[-2]))
    return w_ilc

def get_mcilc_weights(
    inputs: np.ndarray,
    patches: np.ndarray,
    A_cmb: np.ndarray,
    compsep_run: Dict
) -> np.ndarray:
    """
    Compute weights for the MC-ILC method over spatial patches.

    Parameters
    ----------
        inputs : np.ndarray
            Input sky maps, shape (n_channels, n_pixels).
        patches : np.ndarray
            Integer labels assigning each pixel to a spatial patch.
        A_cmb : np.ndarray
            Spectral response of the CMB component, shape (n_channels,).
        compsep_run : dict
            Dictionary with MCILC options, including optional constraints and bias reduction. See 'ilc' for details.

    Returns
    -------
        np.ndarray
            MCILC weights for each pixel, shape (n_channels, n_pixels).
    """
    mask_mcilc = compsep_run.get("mask", np.ones(inputs.shape[1]))

    cov = get_mcilc_cov(inputs, patches, mask_mcilc, reduce_bias=compsep_run["reduce_mcilc_bias"])

    inv_cov = get_inv_cov(cov)
    del cov

    w_mcilc=np.zeros((inputs.shape[0],inputs.shape[1]))

    if "A" in compsep_run:
        compsep_run["A"] = np.vstack((A_cmb, compsep_run["A"]))
        compsep_run["e"] = np.insert(compsep_run["e"], 0, 1.)
        inv_ACA = np.linalg.inv(
            np.einsum("zi,ilk->zlk", compsep_run["A"], np.einsum("ijk,lj->ilk", inv_cov, compsep_run["A"])).T
        ).T
        w_mcilc[:, mask_mcilc > 0.0] = np.einsum(
            "l,ljk->jk", compsep_run["e"],
            np.einsum("lzk,zjk->ljk", inv_ACA, np.einsum("zi,ijk->zjk", compsep_run["A"], inv_cov))
        )
        del inv_ACA
    else:
        AT_invC = np.einsum('j,ijk->ik', A_cmb, inv_cov) # np.sum(inv_cov,axis=1)
        AT_invC_A = np.einsum('j,ijk, i->k', A_cmb, inv_cov, A_cmb) #np.sum(inv_cov,axis=(0,1))
        for i in range(inputs.shape[0]):
            w_mcilc[i, mask_mcilc > 0.] = AT_invC[i]/AT_invC_A
        del AT_invC, AT_invC_A
    
    del inv_cov
    return w_mcilc


def get_ilc_cov(
    input_maps: np.ndarray,
    lmax: int,
    compsep_run: Dict,
    b_ell: np.ndarray
) -> np.ndarray:
    """
    Compute the input covariance matrix for ILC.

    Parameters
    ----------
        input_maps : np.ndarray
            Input sky maps, shape (n_channels, n_pixels).
        lmax : int
            Maximum multipole.
        compsep_run : dict
            Dictionary controlling ILC behavior (bias flags, mask, etc.). See 'ilc' for details.
        b_ell : np.ndarray
            Needlet harmonic window function, if applicable.

    Returns
    -------
        np.ndarray
            Covariance matrix, either global (2D) or pixel-dependent (3D).
    """
    if compsep_run["ilc_bias"] == 0.:
        if "mask" in compsep_run:
            cov=np.mean(np.einsum('ik,jk->ijk', (input_maps * compsep_run["mask"])[:, compsep_run["mask"] > 0.], (input_maps * compsep_run["mask"])[:, compsep_run["mask"] > 0.]),axis=-1)
        else:
            cov=np.mean(np.einsum('ik,jk->ijk', input_maps, input_maps),axis=-1)
    else:
        cov = _get_local_cov(
            input_maps, lmax, compsep_run["ilc_bias"], b_ell=b_ell,
            mask=compsep_run.get("mask"), reduce_bias=compsep_run["reduce_ilc_bias"]
        )
        if "mask" in compsep_run and cov.shape[-1] == input_maps.shape[-1]:
                cov = np.copy(cov[...,compsep_run["mask"] > 0.])
    return cov

def get_mcilc_cov(
    inputs: np.ndarray,
    patches: np.ndarray,
    mask_mcilc: np.ndarray,
    reduce_bias: bool = True,
    mcilc_rings: int = 4
) -> np.ndarray:
    """
    Compute pixel-wise covariance matrices for MC-ILC based on patching and optional donut-masking.

    Parameters
    ----------
        inputs : np.ndarray
            Input sky maps, shape (n_channels, n_pixels).
        patches : np.ndarray
            Pixel-wise patch labels.
        mask_mcilc : np.ndarray
            Binary mask to include/exclude pixels.
        reduce_bias : bool, default=True
            Whether to apply donut masking to reduce MCILC bias.
        mcilc_rings : int, default=4
            Number of HEALPix neighbor rings to exclude when reducing bias.

    Returns
    -------
        np.ndarray
            Covariance matrix per pixel, shape (n_channels, n_channels, n_valid_pixels).
    """
    cov=np.zeros((inputs.shape[0], inputs.shape[0], inputs.shape[-1]))
    
    if reduce_bias:
        neigh = hp.get_all_neighbours(hp.npix2nside(inputs.shape[1]), np.argwhere(mask_mcilc > 0.)[:, 0])
        for pix_ in np.argwhere(mask_mcilc > 0.)[:, 0]:
            donut = np.ones(inputs.shape[1])
            donut[pix_] = 0.   
            if mcilc_rings > 0:
                donut[neigh[:,pix_]]=0.
            if mcilc_rings > 1:
                count=1
                neigh_=neigh[:,pix_]
                while count < mcilc_rings:
                    neigh_=neigh[:,neigh_].flatten()
                    donut[neigh_]=0.
                    count=count+1
            patch_mask = (patches==patches[pix_]) & (donut>0.) & (mask_mcilc>0.)
            cov[...,pix_] = np.cov((inputs * mask_mcilc)[:,patch_mask],rowvar=True)
    else:
        for patch in np.unique(patches):
            patch_mask = (patches == patch) & (mask_mcilc > 0.)
            patch_cov = np.mean(np.einsum('ik,jk->ijk', (inputs * mask_mcilc)[:,patch_mask], (inputs * mask_mcilc)[:,patch_mask]),axis=2)
            #cov[...,(patches==patch)] = np.repeat(patch_cov[:, :, np.newaxis], np.sum(patches==patch), axis=2)
            cov[...,patch_mask] = np.repeat(patch_cov[:, :, np.newaxis], np.sum(patch_mask), axis=2)

    return cov[...,mask_mcilc > 0.]

def get_inv_cov(cov: np.ndarray) -> np.ndarray:
    """
    Invert covariance matrices for ILC or MC-ILC processing.

    Parameters
    ----------
        cov : np.ndarray
            Covariance matrix of shape (n_channels, n_channels) or (n_channels, n_channels, n_pixels).

    Returns
    -------
        np.ndarray
            Inverted covariance matrix with same shape.
    """
    if cov.ndim == 2:
        inv_cov=np.linalg.inv(cov)
    elif cov.ndim == 3:
        inv_cov=np.linalg.inv(cov.T).T
    return inv_cov


__all__ = [
    name
    for name, obj in globals().items()
    if callable(obj) and getattr(obj, "__module__", None) == __name__
]
                    




import numpy as np
import healpy as hp
import mtneedlet as nl
from typing import Dict, Union, List, Tuple, Optional
from .configurations import Configs
from .leakage import purify_master, purify_recycling
from types import SimpleNamespace
import warnings
import sys


def _get_needlet_windows_(needlet_config: Dict, lmax: int) -> np.ndarray:
    """
    Dispatch function to generate needlet windows based on configuration.

    Parameters
    ----------
        needlet_config : dict
            Dictionary containing needlet settings:
            - "needlet_windows": Type of needlet windows ('cosine', 'standard', 'mexican').
            - "ell_peaks": List of integers defining multipoles of peaks for needlet bands (required for 'cosine').
            - "width": Width of the needlet windows (required for 'standard' and 'mexican').
            - "merging_needlets": Integer or list of integers defining ranges of needlets to be merged.
        lmax : int
            Maximum multipole.

    Returns
    -------
        np.ndarray
            Needlet window array of shape (n_bands, lmax+1)
    """
    needlet_options = {
    "cosine": get_cosine_windows,
    'standard': get_standard_windows,
    'mexican': get_mexican_windows}

    if needlet_config["needlet_windows"] not in needlet_options:
        raise ValueError(f"Unsupported needlet type: {needlet_config['needlet_windows']}")

    return needlet_options[needlet_config["needlet_windows"]](lmax, needlet_config)
    
def get_cosine_windows(lmax: int, needlet_config: Dict) -> np.ndarray:
    """
    Generate cosine-based needlet windows.

    Parameters
    ----------
        lmax : int
            Maximum multipole.
        needlet_config : dict
            Configuration with key "ell_peaks" (list of peak positions).

    Returns
    -------
        np.ndarray
            Cosine window array (n_bands, lmax+1).
    """
    if "ell_peaks" not in needlet_config:
        raise ValueError("Cosine needlets require 'ell_peaks' in the config. \
            'ell_peaks' is a list of integers that define the multipoles of the peaks of the cosine windows.")

    def compute_cosine_window(ell_peak: int, ell_ext: int) -> np.ndarray:
        ell_range = np.arange(min(ell_peak, ell_ext), max(ell_peak, ell_ext) + 1)
        return np.cos((ell_peak - ell_range) / (ell_peak - ell_ext) * np.pi / 2.)

    bandpeaks = needlet_config["ell_peaks"]
    n_bands = len(bandpeaks)
    if n_bands < 2:
        warnings.warn("Just one peak provided, no needlets are generated and the method will be performed in pixel domain.",UserWarning)
        return np.ones((1, lmax + 1))
    b_ell = np.zeros((n_bands,lmax+1))

    bandpeaks = [peak for peak in bandpeaks if peak <= lmax]
    bandpeaks = sorted(bandpeaks)

    if bandpeaks[0] > 0:
        b_ell[0,:bandpeaks[0]]=1.

    b_ell[0,bandpeaks[0]:bandpeaks[1]+1] = compute_cosine_window(bandpeaks[0], bandpeaks[1]) #

    if n_bands >= 3:
        for i in range(1, n_bands-1):
            ell_min, ell_p, ell_max = bandpeaks[i - 1], bandpeaks[i], bandpeaks[i + 1]
            b_ell[i,ell_min:ell_p+1] = compute_cosine_window(ell_p, ell_min)
            b_ell[i,ell_p:ell_max+1] = compute_cosine_window(ell_p, ell_max)

    b_ell[-1,bandpeaks[-2]:bandpeaks[-1]+1] = compute_cosine_window(bandpeaks[-1], bandpeaks[-2])
    if bandpeaks[-1] < lmax:
        b_ell[-1,bandpeaks[-1]+1:] = 1.
    return b_ell

def get_standard_windows(lmax: int, needlet_config: Dict) -> np.ndarray:
    """
    Generate standard needlet windows.

    Parameters
    ----------
        lmax : int
            Maximum multipole.
        needlet_config : dict
            Must contain 'width' and optionally 'merging_needlets'.

    Returns
    -------
        np.ndarray
            Standard needlet windows.
    """
    if not "width" in needlet_config:
        raise ValueError("'standard' needlet windows require 'width' parameter. \
            'width' is a floating number that defines the width of the standard needlet windows.")

    j_min, j_max = 0, 2

    while True:
        j_range = list(range(j_min, j_max + 1))
        b_ell = nl.standardneedlet(needlet_config["width"], j_range, lmax)
        if np.abs(np.sum(b_ell**2, axis=0)[-1] - 1.0) <= 1e-5:
            break
        j_max += 1

    return _merge_needlets_(b_ell, needlet_config.get("merging_needlets"))

def get_mexican_windows(lmax: int, needlet_config: Dict) -> np.ndarray:
    """
    Generate Mexican needlet windows.

    Parameters
    ----------
        lmax : int
            Maximum multipole.
        needlet_config : dict
            Must contain 'width' and optionally 'merging_needlets'.

    Returns
    -------
        np.ndarray
            Mexican needlet windows.
    """

    if "width" not in needlet_config:
        raise ValueError("'mexican' needlet windows require 'width' parameter. \
            'width' is a floating number that defines the width of the mexican needlet windows.")

    j_min, j_max = 0, 2
    width = needlet_config["width"]

    while True:
        j_range = list(range(j_min, j_max + 1))
        b_ell = nl.mexicanneedlet(needlet_config["width"], j_range, lmax)
        if np.abs(np.sum(b_ell**2, axis=0)[-1] - 1.0) <= 1e-5:
            break
        j_max += 1
    
    return _merge_needlets_(b_ell, needlet_config.get("merging_needlets"))

    
def _merge_needlets_(b_ell: np.ndarray, merging_needlets: Optional[Union[int, List[int]]]) -> np.ndarray:
    """
    Merge needlet bands according to merging strategy.

    Parameters
    ----------
        b_ell : np.ndarray
            Original needlet windows with shape (n_bands, lmax+1).
        merging_needlets : int or list of int or None
            Strategy for merging bands. If None no merging is performed.
            If int, merges first `merging_needlets` bands into one.
            If list, merges bands according to the ranges defined in the list.

    Returns
    -------
        np.ndarray
            Merged needlet windows.
    """
    if merging_needlets is None:
        return b_ell

    n_bands, lmaxp1 = b_ell.shape

    if isinstance(merging_needlets, int):
        if n_bands <= merging_needlets:
            warnings.warn("Number of needlets to merge in the first band is larger than the number of needlets. \
                All needlets are merged in one band. You are therefore running in pixel domain.",UserWarning)
            return np.sqrt(np.sum(b_ell**2, axis=0))[None, :]
        merged_bell = np.sqrt(np.sum((b_ell**2)[:merging_needlets],axis=0))[None, :]
        return np.concatenate([merged_bell, b_ell[merging_needlets:]])

    elif isinstance(merging_needlets, list):
        merging_needlets = sorted(set(merging_needlets))

        if b_ell.shape[0] <= merging_needlets[0]:
            warnings.warn("Number of needlets to merge in the first band is larger than the number of needlets. \
                All needlets are therefore merged in one band and the method will be performed in pixel domain.",UserWarning)
            return  np.sqrt(np.sum(b_ell**2, axis=0))[None, :]

        if merging_needlets[0] > 0:
            merging_needlets = [0] + merging_needlets
        if merging_needlets[-1] >= n_bands:
            merging_needlets = [x for x in merging_needlets if x < n_bands]
            merging_needlets.append(n_bands)

        merged_b_ell = []
        for start, end in zip(merging_needlets[:-1], merging_needlets[1:]):
            merged_b_ell.append(np.sqrt(np.sum((b_ell**2)[start:end], axis=0)))
        merged_b_ell = np.array(merged_b_ell)
        if merging_needlets[-1] < n_bands:
            merged_b_ell = np.concatenate([merged_b_ell, b_ell[merging_needlets[-1]:]])
        return merged_b_ell
    else:
        raise ValueError("merging_needlets must be int, list[int], or None")

def _get_nside_lmax_from_b_ell(b_ell: np.ndarray, nside: int, lmax: int) -> Tuple[int, int]:
    """
    Estimate optimal nside and lmax based on the needlet filter support.

    Parameters
    ----------
        b_ell : np.ndarray
            Needlet filters of shape (n_bands, lmax+1).
        nside : int
            Input nside.
        lmax : int
            Input lmax.

    Returns
    -------
        Tuple[int, int]
            Recommended (nside, lmax) values to be used for needlet maps generation.
    """

    max_b = np.max(np.nonzero(b_ell)) 
    if max_b == lmax:
        return nside, lmax
    
    max_b = max_b / 2

    # If the band is non zero only for ell <= 16, nside and lmax are set to 8 and 16
    if max_b <= 8:
        return 8, 16
    
    for k in range(3, 12):
        lower, upper = 2**k, 2**(k + 1)
        if lower < max_b <= upper:
            if upper <= nside:
                return upper, 2 * upper
            else:
                return nside, 2 * nside
    return nside, lmax

def _needlet_filtering(alms: np.ndarray, b_ell: np.ndarray, lmax_out: int) -> np.ndarray:
    """
    Filter alm coefficients using a given needlet filter.

    Parameters
    ----------
        alms : np.ndarray
            Input spherical harmonic coefficients of shape (n_alms, n_components) or (n_alms,).
        b_ell : np.ndarray
            Needlet filter of length lmax+1.
        lmax_out : int
            Desired output lmax.

    Returns
    -------
        np.ndarray
            Filtered and possibly resized alm coefficients.
    """
    if alms.ndim == 2:
        filtered_alms = np.array([hp.almxfl(alms[:,c], b_ell) for c in range(alms.shape[-1])]).T
    elif alms.ndim == 1:
        filtered_alms = hp.almxfl(alms, b_ell)

    if alms.shape[0] == hp.Alm.getsize(lmax_out):
        return filtered_alms
    else:
        lmax_j = min(hp.Alm.getlmax(alms.shape[0]), lmax_out)
        idx_lmax_in = np.array([hp.Alm.getidx(hp.Alm.getlmax(filtered_alms.shape[0]), ell, m) for ell in range(0, lmax_j+1) for m in range(ell + 1)])
        idx_lmax_out = np.array([hp.Alm.getidx(lmax_out, ell, m) for ell in range(0, lmax_j+1) for m in range(ell + 1)])
        if alms.ndim == 2:
            alms_j = np.zeros((hp.Alm.getsize(lmax_out), alms.shape[-1]), dtype=complex)
            alms_j[idx_lmax_out, :] = filtered_alms[idx_lmax_in, :]
        elif alms.ndim == 1:
            alms_j = np.zeros((hp.Alm.getsize(lmax_out)), dtype=complex)
            alms_j[idx_lmax_out] = filtered_alms[idx_lmax_in]
        return alms_j

def _get_good_channels_nl(config: Configs, b_ell: np.ndarray, threshold: float = 1e-5) -> np.ndarray:
    """
    Returns indices of frequency channels with acceptable beam transfer function ratios.

    Parameters
    ----------
        config : Configs
            Configuration object with instrument parameters.
        b_ell : np.ndarray
            Needlet harmonic window.
        threshold : float
            Minimum acceptable ratio of input/output beam values.

    Returns
    -------
        np.ndarray
            Indices of "good" channels.
    """
    freqs_nl = []
    for i, fwhm in enumerate(config.instrument.fwhm):
        bl_in = hp.gauss_beam(np.radians(fwhm/60.), lmax=config.lmax,pol=False)
        bl_out = hp.gauss_beam(np.radians(config.fwhm_out/60.), lmax=config.lmax,pol=False)
        if not np.any((bl_in / bl_out)[b_ell > 1e-2] < threshold):
            freqs_nl.append(i)
    return np.array(freqs_nl)

__all__ = [
    name
    for name, obj in globals().items()
    if callable(obj) and getattr(obj, "__module__", None) == __name__
]
                    
